[CCCL] 瘦身 + 补全: 移除 cudax/python/libcudacxx-tests 冗余文件, 新增 c2h 测试助手 + cmake 构建系统 + 8 个 CUDA thrust examples
变更摘要:
- 删除: cudax/ (783 files, 7.2M) — 实验性组件,竞赛不需要
- 删除: python/ (226 files, 2.0M) — Python 绑定,竞赛不需要
- 删除: libcudacxx/{test,benchmarks,codegen,cmake,share} (4432 files, 31M)
保留: libcudacxx/include/ (1463 headers, cuda::std 编译依赖)
- 新增: c2h/ (27 files) — CUB Catch2 测试辅助头文件,编译 243 个测试必需
- 新增: cmake/ (29 files) — CCCL 原生 CMake 构建系统
- 新增: thrust/examples/cuda/ (7 files) + cpp_integration/ (1 file)
async_reduce, custom_temporary_allocation, explicit_cuda_stream,
global_device_vector, range_view, unwrap_pointer, wrap_pointer, device
结果: cccl_upstream 从 74M→35M (瘦身 53%), 核心内容 100% 保留:
27/27 tuning headers, 78 benchmarks, 243 tests,
60 thrust examples, 18 CUB examples, 全部编译头文件
This commit is contained in:
@@ -1,75 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/__places/partitions/blocked_partition.cuh>
|
||||
#include <cuda/experimental/__places/partitions/cyclic_shape.cuh>
|
||||
#include <cuda/experimental/stf.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
double X0(int i)
|
||||
{
|
||||
return sin((double) i);
|
||||
}
|
||||
|
||||
double Y0(int i)
|
||||
{
|
||||
return cos((double) i);
|
||||
}
|
||||
|
||||
int main()
|
||||
{
|
||||
stream_ctx ctx;
|
||||
|
||||
const int N = 128;
|
||||
double X[N], Y[N];
|
||||
|
||||
for (int ind = 0; ind < N; ind++)
|
||||
{
|
||||
X[ind] = X0(ind);
|
||||
Y[ind] = Y0(ind);
|
||||
}
|
||||
|
||||
const double alpha = 3.14;
|
||||
|
||||
auto handle_X = ctx.logical_data(X, {N});
|
||||
auto handle_Y = ctx.logical_data(Y, {N});
|
||||
|
||||
auto number_devices = 4;
|
||||
auto all_devs = exec_place::repeat(exec_place::device(0), number_devices);
|
||||
|
||||
auto spec = par(16 * 4, par(4));
|
||||
ctx.launch(spec, all_devs, handle_X.read(), handle_Y.rw())->*[=] _CCCL_DEVICE(auto th, auto x, auto y) {
|
||||
// Blocked partition among elements in the outer most level
|
||||
auto outer_sh = blocked_partition::apply(shape(x), pos4(th.rank(0)), dim4(th.size(0)));
|
||||
|
||||
// Cyclic partition among elements in the remaining levels
|
||||
auto inner_sh = cyclic_partition::apply(outer_sh, pos4(th.inner().rank()), dim4(th.inner().size()));
|
||||
|
||||
for (auto ind : inner_sh)
|
||||
{
|
||||
y(ind) += alpha * x(ind);
|
||||
}
|
||||
};
|
||||
|
||||
ctx.host_launch(handle_X.read(), handle_Y.read())->*[=](auto X, auto Y) {
|
||||
for (int ind = 0; ind < N; ind++)
|
||||
{
|
||||
// Y should be Y0 + alpha X0
|
||||
// fprintf(stderr, "Y[%ld] = %lf - expect %lf\n", ind, Y(ind), (Y0(ind) + alpha * X0(ind)));
|
||||
EXPECT(fabs(Y(ind) - (Y0(ind) + alpha * X0(ind))) < 0.0001);
|
||||
|
||||
// X should be X0
|
||||
EXPECT(fabs(X(ind) - X0(ind)) < 0.0001);
|
||||
}
|
||||
};
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,119 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
/**
|
||||
* @file
|
||||
*
|
||||
* @brief This example illustrates how to use the task construct with grids of
|
||||
* places and composite data places
|
||||
*/
|
||||
|
||||
#include <cuda/experimental/__places/partitions/tiled_partition.cuh>
|
||||
#include <cuda/experimental/__stf/graph/graph_ctx.cuh>
|
||||
#include <cuda/experimental/__stf/stream/stream_ctx.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
template <typename T>
|
||||
__global__ void axpy(size_t start, size_t cnt, T a, const T* x, T* y)
|
||||
{
|
||||
int tid = blockIdx.x * blockDim.x + threadIdx.x;
|
||||
int nthreads = gridDim.x * blockDim.x;
|
||||
|
||||
for (int ind = tid; ind < cnt; ind += nthreads)
|
||||
{
|
||||
y[ind + start] += a * x[ind + start];
|
||||
}
|
||||
}
|
||||
|
||||
double X0(size_t i)
|
||||
{
|
||||
return sin((double) i);
|
||||
}
|
||||
|
||||
double Y0(size_t i)
|
||||
{
|
||||
return cos((double) i);
|
||||
}
|
||||
|
||||
template <typename Ctx>
|
||||
void run()
|
||||
{
|
||||
Ctx ctx;
|
||||
|
||||
const int N = 1024 * 1024 * 32;
|
||||
double *X, *Y;
|
||||
|
||||
X = new double[N];
|
||||
Y = new double[N];
|
||||
SCOPE(exit)
|
||||
{
|
||||
delete[] X;
|
||||
delete[] Y;
|
||||
};
|
||||
|
||||
for (size_t ind = 0; ind < N; ind++)
|
||||
{
|
||||
X[ind] = X0(ind);
|
||||
Y[ind] = Y0(ind);
|
||||
}
|
||||
|
||||
// std::shared_ptr<execution_grid> all_devs = exec_place::all_devices();
|
||||
// use grid [ 0 0 0 0 ] for debugging purpose
|
||||
auto all_devs = exec_place::repeat(exec_place::device(0), 4);
|
||||
|
||||
// 512k doubles = 4MB (2 pages)
|
||||
// A 1D blocking strategy over all devices with a block size of 32 and a round robin distribution of blocks across
|
||||
// devices
|
||||
// data_place cdp = data_place(exec_place::all_devices().as_grid().get_grid(),
|
||||
// [](dim4 grid_dim, pos4 index_pos) { return pos4((index_pos.x / (512 * 1024ULL)) % grid_dim.x); });
|
||||
|
||||
data_place cdp = data_place::composite(tiled_partition<512 * 1024ULL>(), all_devs);
|
||||
|
||||
auto handle_X = ctx.logical_data(X, {N});
|
||||
auto handle_Y = ctx.logical_data(Y, {N});
|
||||
|
||||
double alpha = 3.14;
|
||||
|
||||
/* Compute Y = Y + alpha X */
|
||||
auto t = ctx.task(all_devs, handle_X.read(cdp), handle_Y.rw(cdp));
|
||||
t->*[&](auto, auto sX, auto sY) {
|
||||
size_t grid_size = t.grid_dims().size();
|
||||
|
||||
assert(N % grid_size == 0);
|
||||
|
||||
for (size_t i = 0; i < grid_size; i++)
|
||||
{
|
||||
auto active = t.activate_place(i);
|
||||
axpy<<<16, 128, 0, t.get_stream(i)>>>(i * N / grid_size, N / grid_size, alpha, sX.data_handle(), sY.data_handle());
|
||||
}
|
||||
};
|
||||
|
||||
/* Check the result on the host */
|
||||
ctx.host_launch(handle_X.read(), handle_Y.read())->*[&](auto sX, auto sY) {
|
||||
for (size_t ind = 0; ind < N; ind++)
|
||||
{
|
||||
// Y should be Y0 + alpha X0
|
||||
EXPECT(fabs(sY(ind) - (Y0(ind) + alpha * X0(ind))) < 0.0001);
|
||||
|
||||
// X should be X0
|
||||
EXPECT(fabs(sX(ind) - X0(ind)) < 0.0001);
|
||||
}
|
||||
};
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
|
||||
int main()
|
||||
{
|
||||
run<stream_ctx>();
|
||||
// Disabled until composite data places are implemented with graphs
|
||||
// run<graph_ctx>();
|
||||
}
|
||||
@@ -1,256 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/__stf/stream/stream_ctx.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
/*
|
||||
* DATA BLOCKS
|
||||
* | GHOSTS | DATA | GHOSTS |
|
||||
*/
|
||||
template <typename T>
|
||||
class data_block
|
||||
{
|
||||
public:
|
||||
data_block(stream_ctx& ctx, size_t beg, size_t end, size_t GHOST_SIZE)
|
||||
: beg(beg)
|
||||
, end(end)
|
||||
, block_size(end - beg)
|
||||
, ghost_size(GHOST_SIZE)
|
||||
, array(std::vector<T>(block_size + 2 * ghost_size))
|
||||
, handle(ctx.logical_data(&array[0], block_size + 2 * ghost_size))
|
||||
{}
|
||||
|
||||
public:
|
||||
size_t beg;
|
||||
size_t end;
|
||||
size_t block_size;
|
||||
size_t ghost_size;
|
||||
int dev_id;
|
||||
|
||||
private:
|
||||
std::vector<T> array;
|
||||
|
||||
public:
|
||||
// HANDLE = whole data + boundaries
|
||||
logical_data<slice<T>> handle;
|
||||
};
|
||||
|
||||
template <typename T>
|
||||
T check_sum(stream_ctx& ctx, data_block<T>& bn)
|
||||
{
|
||||
T sum = 0.0;
|
||||
|
||||
auto t = ctx.task(exec_place::host(), bn.handle.read());
|
||||
t->*[&](cudaStream_t stream, auto h_center) {
|
||||
cuda_safe_call(cudaStreamSynchronize(stream));
|
||||
for (size_t offset = bn.ghost_size; offset < bn.ghost_size + bn.block_size; offset++)
|
||||
{
|
||||
sum += h_center.data_handle()[offset];
|
||||
}
|
||||
};
|
||||
|
||||
return sum;
|
||||
}
|
||||
|
||||
// array and array1 have a size of (cnt + 2*ghost_size)
|
||||
template <typename T>
|
||||
__global__ void stencil_kernel(size_t cnt, size_t ghost_size, T* array, const T* array1)
|
||||
{
|
||||
for (size_t idx = threadIdx.x + blockIdx.x * blockDim.x; idx < cnt; idx += blockDim.x * gridDim.x)
|
||||
{
|
||||
size_t idx2 = idx + ghost_size;
|
||||
array[idx2] = 0.9 * array1[idx2] + 0.05 * array1[idx2 - 1] + 0.05 * array1[idx2 + 1];
|
||||
}
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
void stencil(stream_ctx& ctx, data_block<T>& bn, data_block<T>& bn1)
|
||||
{
|
||||
int dev = bn.dev_id;
|
||||
|
||||
auto t = ctx.task(exec_place::device(dev), bn.handle.rw(), bn1.handle.read());
|
||||
t->*[&](cudaStream_t stream, auto bn_array, auto bn1_array) {
|
||||
stencil_kernel<T>
|
||||
<<<32, 64, 0, stream>>>(bn.block_size, bn.ghost_size, bn_array.data_handle(), bn1_array.data_handle());
|
||||
};
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
__global__ void copy_kernel(size_t cnt, T* dst, const T* src)
|
||||
{
|
||||
for (size_t idx = threadIdx.x + blockIdx.x * blockDim.x; idx < cnt; idx += blockDim.x * gridDim.x)
|
||||
{
|
||||
dst[idx] = src[idx];
|
||||
}
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
void copy_task(
|
||||
stream_ctx& ctx,
|
||||
size_t cnt,
|
||||
logical_data<slice<T>>& dst,
|
||||
size_t offset_dst,
|
||||
int dst_dev,
|
||||
logical_data<slice<T>>& src,
|
||||
size_t offset_src,
|
||||
int src_dev)
|
||||
{
|
||||
auto t = ctx.task(exec_place::device(dst_dev), dst.rw(), src.read(data_place::device(src_dev)));
|
||||
t->*[&](cudaStream_t stream, auto dst_array, auto src_array) {
|
||||
int nblocks = (cnt > 64) ? 32 : 1;
|
||||
copy_kernel<T>
|
||||
<<<nblocks, 64, 0, stream>>>(cnt, dst_array.data_handle() + offset_dst, src_array.data_handle() + offset_src);
|
||||
};
|
||||
}
|
||||
|
||||
// Copy left/right handles from neighbours to the array
|
||||
template <typename T>
|
||||
void update_halo(stream_ctx& ctx, data_block<T>& bn, data_block<T>& left, data_block<T>& right)
|
||||
{
|
||||
size_t gs = bn.ghost_size;
|
||||
size_t bs = bn.block_size;
|
||||
|
||||
// Copy the bn.ghost_size last computed items in "left" (outside the halo)
|
||||
copy_task<T>(ctx, gs, bn.handle, 0, bn.dev_id, left.handle, bs, left.dev_id);
|
||||
|
||||
// Copy the bn.ghost_size first computed items (outside the halo)
|
||||
copy_task<T>(ctx, gs, bn.handle, gs + bs, bn.dev_id, right.handle, gs, right.dev_id);
|
||||
}
|
||||
|
||||
// Copy inner part of bn into bn1
|
||||
template <typename T>
|
||||
void copy_inner(stream_ctx& ctx, data_block<T>& bn1, data_block<T>& bn)
|
||||
{
|
||||
size_t gs = bn.ghost_size;
|
||||
size_t bs = bn.block_size;
|
||||
|
||||
int dev_id = bn.dev_id;
|
||||
|
||||
// Copy the bn.ghost_size last computed items in "left" (outside the halo)
|
||||
copy_task<T>(ctx, bs, bn1.handle, gs, dev_id, bn.handle, gs, dev_id);
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
int ndevs;
|
||||
cuda_safe_call(cudaGetDeviceCount(&ndevs));
|
||||
|
||||
stream_ctx ctx;
|
||||
|
||||
int NITER = 500;
|
||||
size_t NBLOCKS = 4 * ndevs;
|
||||
size_t BLOCK_SIZE = 2048 * 1024;
|
||||
|
||||
if (argc > 1)
|
||||
{
|
||||
NITER = atoi(argv[1]);
|
||||
}
|
||||
|
||||
if (argc > 2)
|
||||
{
|
||||
NBLOCKS = atoi(argv[2]);
|
||||
}
|
||||
|
||||
const size_t GHOST_SIZE = 1;
|
||||
|
||||
size_t TOTAL_SIZE = NBLOCKS * BLOCK_SIZE;
|
||||
|
||||
double* U0 = new double[NBLOCKS * BLOCK_SIZE];
|
||||
for (size_t idx = 0; idx < NBLOCKS * BLOCK_SIZE; idx++)
|
||||
{
|
||||
U0[idx] = (idx == 0) ? 1.0 : 0.0;
|
||||
}
|
||||
|
||||
std::vector<data_block<double>> Un;
|
||||
std::vector<data_block<double>> Un1;
|
||||
|
||||
// Create blocks and allocates host data
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
size_t beg = b * BLOCK_SIZE;
|
||||
size_t end = (b + 1) * BLOCK_SIZE;
|
||||
|
||||
Un.emplace_back(ctx, beg, end, 1ull);
|
||||
Un1.emplace_back(ctx, beg, end, 1ull);
|
||||
}
|
||||
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
Un[b].dev_id = b % ndevs;
|
||||
Un1[b].dev_id = b % ndevs;
|
||||
}
|
||||
|
||||
// Fill blocks with initial values. For the sake of simplicity, we are
|
||||
// using a synchronization primitive and host code, but this could have
|
||||
// been written asynchronously using host callbacks.
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
size_t beg = b * BLOCK_SIZE;
|
||||
|
||||
auto t = ctx.task(exec_place::host(), Un[b].handle.rw(), Un1[b].handle.rw());
|
||||
t->*[&](cudaStream_t stream, auto Un_vals, auto Un1_vals) {
|
||||
cuda_safe_call(cudaStreamSynchronize(stream));
|
||||
for (size_t local_idx = 0; local_idx < BLOCK_SIZE; local_idx++)
|
||||
{
|
||||
double val = U0[(beg + local_idx + TOTAL_SIZE) % TOTAL_SIZE];
|
||||
Un1_vals.data_handle()[local_idx + GHOST_SIZE] = val;
|
||||
Un_vals.data_handle()[local_idx + GHOST_SIZE] = val;
|
||||
}
|
||||
};
|
||||
}
|
||||
|
||||
for (int iter = 0; iter < NITER; iter++)
|
||||
{
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
update_halo(ctx, Un1[b], Un[(b - 1 + NBLOCKS) % NBLOCKS], Un[(b + 1) % NBLOCKS]);
|
||||
}
|
||||
|
||||
// UPDATE Un from Un1
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
stencil(ctx, Un[b], Un1[b]);
|
||||
}
|
||||
|
||||
#if 0
|
||||
// We make sure that the total sum of elements remains constant
|
||||
if (iter % 250 == 0)
|
||||
{
|
||||
double sum = 0.0;
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
sum += check_sum(ctx, Un[b]);
|
||||
}
|
||||
|
||||
// fprintf(stderr, "iter %d : CHECK SUM = %e\n", iter, sum);
|
||||
}
|
||||
#endif
|
||||
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
// Copy inner part of Un into Un1
|
||||
copy_inner(ctx, Un[b], Un1[b]);
|
||||
}
|
||||
}
|
||||
|
||||
// In this stencil, the sum of the elements is supposed to be a constant
|
||||
double sum = 0.0;
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
sum += check_sum(ctx, Un[b]);
|
||||
}
|
||||
|
||||
double err = fabs(sum - 1.0);
|
||||
EXPECT(err < 0.0001);
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,92 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/__places/partitions/tiled_partition.cuh>
|
||||
#include <cuda/experimental/__stf/stream/stream_ctx.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
template <typename T>
|
||||
__global__ void stencil_kernel(slice<T> Un, slice<const T> Un1)
|
||||
{
|
||||
size_t N = Un.extent(0);
|
||||
for (size_t i = threadIdx.x + blockIdx.x * blockDim.x; i < N; i += blockDim.x * gridDim.x)
|
||||
{
|
||||
Un(i) = 0.9 * Un1(i) + 0.05 * Un1((i + N - 1) % N) + 0.05 * Un1((i + 1) % N);
|
||||
}
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
stream_ctx ctx;
|
||||
|
||||
int NITER = 500;
|
||||
int NBLOCKS = 20;
|
||||
const size_t BLOCK_SIZE = 2048 * 1024;
|
||||
|
||||
if (argc > 1)
|
||||
{
|
||||
NITER = atoi(argv[1]);
|
||||
}
|
||||
|
||||
if (argc > 2)
|
||||
{
|
||||
NBLOCKS = atoi(argv[2]);
|
||||
}
|
||||
|
||||
const size_t TOTAL_SIZE = NBLOCKS * BLOCK_SIZE;
|
||||
|
||||
double* Un = new double[TOTAL_SIZE];
|
||||
double* Un1 = new double[TOTAL_SIZE];
|
||||
|
||||
for (size_t idx = 0; idx < TOTAL_SIZE; idx++)
|
||||
{
|
||||
Un[idx] = (idx == 0) ? 1.0 : 0.0;
|
||||
Un1[idx] = Un[idx];
|
||||
}
|
||||
|
||||
auto lUn = ctx.logical_data(make_slice(Un, TOTAL_SIZE));
|
||||
auto lUn1 = ctx.logical_data(make_slice(Un1, TOTAL_SIZE));
|
||||
|
||||
// std::shared_ptr<execution_grid> all_devs = exec_place::all_devices();
|
||||
// use grid [ 0 0 0 0 ] for debugging purpose
|
||||
auto all_devs = exec_place::repeat(exec_place::device(0), 4);
|
||||
|
||||
data_place cdp = data_place::composite(tiled_partition<BLOCK_SIZE>(), all_devs);
|
||||
|
||||
for (int iter = 0; iter < NITER; iter++)
|
||||
{
|
||||
// UPDATE Un from Un1
|
||||
ctx.task(lUn.rw(cdp), lUn1.read(cdp))->*[&](auto stream, auto sUn, auto sUn1) {
|
||||
stencil_kernel<double><<<32, 128, 0, stream>>>(sUn, sUn1);
|
||||
};
|
||||
|
||||
// We make sure that the total sum of elements remains constant
|
||||
if (iter % 250 == 0)
|
||||
{
|
||||
double sum = 0.0;
|
||||
|
||||
ctx.task(exec_place::host(), lUn.read())->*[&](auto stream, auto sUn) {
|
||||
cuda_safe_call(cudaStreamSynchronize(stream));
|
||||
for (size_t offset = 0; offset < TOTAL_SIZE; offset++)
|
||||
{
|
||||
sum += sUn(offset);
|
||||
}
|
||||
};
|
||||
|
||||
// TODO add an assertion to check whether sum is close enough to 1.0
|
||||
// fprintf(stderr, "iter %d : CHECK SUM = %e\n", iter, sum);
|
||||
}
|
||||
|
||||
std::swap(lUn, lUn1);
|
||||
}
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,265 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/__stf/stream/stream_ctx.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
static stream_ctx ctx;
|
||||
|
||||
/*
|
||||
* DATA BLOCKS
|
||||
* | GHOSTS | DATA | GHOSTS |
|
||||
*/
|
||||
template <typename T>
|
||||
class data_block
|
||||
{
|
||||
public:
|
||||
data_block(size_t beg, size_t end, size_t GHOST_SIZE)
|
||||
: beg(beg)
|
||||
, end(end)
|
||||
, block_size(end - beg)
|
||||
, ghost_size(GHOST_SIZE)
|
||||
, array(std::vector<T>(block_size + 2 * ghost_size))
|
||||
, left_interface(std::vector<T>(ghost_size))
|
||||
, right_interface(std::vector<T>(ghost_size))
|
||||
, handle(ctx.logical_data(&array[0], block_size + 2 * ghost_size))
|
||||
, left_handle(ctx.logical_data(&left_interface[0], ghost_size))
|
||||
, right_handle(ctx.logical_data(&right_interface[0], ghost_size))
|
||||
{}
|
||||
|
||||
T check_sum()
|
||||
{
|
||||
T sum = 0.0;
|
||||
|
||||
ctx.task(exec_place::host(), handle.read())->*[&](cudaStream_t stream, auto sn) {
|
||||
cuda_safe_call(cudaStreamSynchronize(stream));
|
||||
const T* h_center = sn.data_handle();
|
||||
for (size_t offset = ghost_size; offset < ghost_size + block_size; offset++)
|
||||
{
|
||||
sum += h_center[offset];
|
||||
}
|
||||
};
|
||||
|
||||
return sum;
|
||||
}
|
||||
|
||||
public:
|
||||
size_t beg;
|
||||
size_t end;
|
||||
size_t block_size;
|
||||
size_t ghost_size;
|
||||
int preferred_device;
|
||||
|
||||
private:
|
||||
std::vector<T> array;
|
||||
std::vector<T> left_interface;
|
||||
std::vector<T> right_interface;
|
||||
|
||||
public:
|
||||
// HANDLE = whole data + boundaries
|
||||
logical_data<slice<T>> handle;
|
||||
// A piece of data to store the left part of the block
|
||||
logical_data<slice<T>> left_handle;
|
||||
// A piece of data to store the right part of the block
|
||||
logical_data<slice<T>> right_handle;
|
||||
};
|
||||
|
||||
// array and array1 have a size of (cnt + 2*ghost_size)
|
||||
template <typename T>
|
||||
__global__ void stencil_kernel(size_t cnt, size_t ghost_size, T* array, const T* array1)
|
||||
{
|
||||
for (size_t idx = threadIdx.x + blockIdx.x * blockDim.x; idx < cnt; idx += blockDim.x * gridDim.x)
|
||||
{
|
||||
size_t idx2 = idx + ghost_size;
|
||||
array[idx2] = 0.9 * array1[idx2] + 0.05 * array1[idx2 - 1] + 0.05 * array1[idx2 + 1];
|
||||
}
|
||||
}
|
||||
|
||||
// bn1.array = bn.array
|
||||
template <typename T>
|
||||
void stencil(data_block<T>& bn, data_block<T>& bn1)
|
||||
{
|
||||
int dev = bn.preferred_device;
|
||||
|
||||
ctx.task(exec_place::device(dev), bn.handle.rw(), bn1.handle.read())->*[&](cudaStream_t stream, auto sN, auto sN1) {
|
||||
stencil_kernel<T><<<32, 64, 0, stream>>>(bn.block_size, bn.ghost_size, sN.data_handle(), sN1.data_handle());
|
||||
};
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
__global__ void copy_kernel(size_t cnt, T* dst, const T* src)
|
||||
{
|
||||
for (size_t idx = threadIdx.x + blockIdx.x * blockDim.x; idx < cnt; idx += blockDim.x * gridDim.x)
|
||||
{
|
||||
dst[idx] = src[idx];
|
||||
}
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
void copy_task(
|
||||
size_t cnt, logical_data<slice<T>>& dst, size_t offset_dst, logical_data<slice<T>>& src, size_t offset_src, int dev)
|
||||
{
|
||||
ctx.task(exec_place::device(dev), dst.rw(), src.read())->*[&](cudaStream_t stream, auto dstS, auto srcS) {
|
||||
int nblocks = (cnt > 64) ? 32 : 1;
|
||||
copy_kernel<T><<<nblocks, 64, 0, stream>>>(cnt, dstS.data_handle() + offset_dst, srcS.data_handle() + offset_src);
|
||||
};
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
void update_inner_interfaces(data_block<T>& bn)
|
||||
{
|
||||
// LEFT
|
||||
copy_task<T>(bn.ghost_size, bn.left_handle, 0, bn.handle, bn.ghost_size, bn.preferred_device);
|
||||
|
||||
// RIGHT
|
||||
copy_task<T>(bn.ghost_size, bn.right_handle, 0, bn.handle, bn.block_size, bn.preferred_device);
|
||||
}
|
||||
|
||||
// Copy left/right handles from neighbours to the array
|
||||
template <typename T>
|
||||
void update_outer_interfaces(data_block<T>& bn, data_block<T>& left, data_block<T>& right)
|
||||
{
|
||||
// update_outer_interface_left
|
||||
copy_task<T>(bn.ghost_size, bn.handle, 0, left.right_handle, 0, bn.preferred_device);
|
||||
|
||||
// update_outer_interface_right
|
||||
copy_task<T>(bn.ghost_size, bn.handle, bn.ghost_size + bn.block_size, right.left_handle, 0, bn.preferred_device);
|
||||
}
|
||||
|
||||
// bn1.array = bn.array
|
||||
template <typename T>
|
||||
void copy_array(data_block<T>& bn, data_block<T>& bn1)
|
||||
{
|
||||
assert(bn.preferred_device == bn1.preferred_device);
|
||||
copy_task<T>(bn.block_size + 2 * bn.ghost_size, bn1.handle, 0, bn.handle, 0, bn.preferred_device);
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
int NITER = 500;
|
||||
size_t NBLOCKS = 4;
|
||||
size_t BLOCK_SIZE = 2048 * 1024;
|
||||
|
||||
if (argc > 1)
|
||||
{
|
||||
NITER = atoi(argv[1]);
|
||||
}
|
||||
|
||||
if (argc > 2)
|
||||
{
|
||||
NBLOCKS = atoi(argv[2]);
|
||||
}
|
||||
|
||||
const size_t GHOST_SIZE = 1;
|
||||
|
||||
size_t TOTAL_SIZE = NBLOCKS * BLOCK_SIZE;
|
||||
|
||||
int ndevs;
|
||||
cuda_safe_call(cudaGetDeviceCount(&ndevs));
|
||||
|
||||
// fprintf(stderr, "GOT %d devices\n", ndevs);
|
||||
|
||||
double* U0 = new double[NBLOCKS * BLOCK_SIZE];
|
||||
for (size_t idx = 0; idx < NBLOCKS * BLOCK_SIZE; idx++)
|
||||
{
|
||||
U0[idx] = (idx == 0) ? 1.0 : 0.0;
|
||||
}
|
||||
|
||||
std::vector<data_block<double>> Un;
|
||||
std::vector<data_block<double>> Un1;
|
||||
|
||||
// Create blocks and allocates host data
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
size_t beg = b * BLOCK_SIZE;
|
||||
size_t end = (b + 1) * BLOCK_SIZE;
|
||||
|
||||
Un.emplace_back(beg, end, 1ull);
|
||||
Un1.emplace_back(beg, end, 1ull);
|
||||
}
|
||||
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
Un[b].preferred_device = b % ndevs;
|
||||
Un1[b].preferred_device = b % ndevs;
|
||||
}
|
||||
|
||||
// Fill blocks with initial values. For the sake of simplicity, we are
|
||||
// using a synchronization primitive and host code, but this could have
|
||||
// been written asynchronously using host callbacks.
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
size_t beg = b * BLOCK_SIZE;
|
||||
|
||||
ctx.task(exec_place::host(), Un1[b].handle.rw())->*[&](cudaStream_t stream, auto sUn1) {
|
||||
cuda_safe_call(cudaStreamSynchronize(stream));
|
||||
double* Un1_vals = sUn1.data_handle();
|
||||
|
||||
for (size_t local_idx = 0; local_idx < BLOCK_SIZE; local_idx++)
|
||||
{
|
||||
Un1_vals[local_idx + GHOST_SIZE] = U0[(beg + local_idx + TOTAL_SIZE) % TOTAL_SIZE];
|
||||
}
|
||||
};
|
||||
}
|
||||
|
||||
for (int iter = 0; iter < NITER; iter++)
|
||||
{
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
// Update the internal copies of the left and right boundaries
|
||||
update_inner_interfaces(Un1[b]);
|
||||
}
|
||||
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
// Apply ghost cells from neighbours to put then in the "center" array
|
||||
update_outer_interfaces(Un1[b], Un1[(b - 1 + NBLOCKS) % NBLOCKS], Un1[(b + 1) % NBLOCKS]);
|
||||
}
|
||||
|
||||
// UPDATE Un from Un1
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
stencil(Un[b], Un1[b]);
|
||||
}
|
||||
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
// Save Un into Un1
|
||||
copy_array(Un[b], Un1[b]);
|
||||
}
|
||||
|
||||
#if 0
|
||||
// We make sure that the total sum of elements remains constant
|
||||
if (iter % 250 == 0)
|
||||
{
|
||||
double check_sum = 0.0;
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
check_sum += Un[b].check_sum();
|
||||
}
|
||||
|
||||
// fprintf(stderr, "iter %d : CHECK SUM = %e\n", iter, check_sum);
|
||||
}
|
||||
#endif
|
||||
}
|
||||
|
||||
// In this stencil, the sum of the elements is supposed to be a constant
|
||||
double check_sum = 0.0;
|
||||
for (size_t b = 0; b < NBLOCKS; b++)
|
||||
{
|
||||
check_sum += Un[b].check_sum();
|
||||
}
|
||||
|
||||
double err = fabs(check_sum - 1.0);
|
||||
EXPECT(err < 0.0001);
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,114 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/__places/partitions/tiled_partition.cuh>
|
||||
#include <cuda/experimental/__stf/stream/stream_ctx.cuh>
|
||||
#include <cuda/experimental/__stf/utility/pretty_print.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
template <typename T>
|
||||
__global__ void stencil2D_kernel(slice<T, 2> sUn, slice<const T, 2> sUn1)
|
||||
{
|
||||
size_t N = sUn.extent(0);
|
||||
for (size_t i = threadIdx.x + blockIdx.x * blockDim.x; i < N; i += blockDim.x * gridDim.x)
|
||||
{
|
||||
for (size_t j = 0; j < N; j++)
|
||||
{
|
||||
sUn(j, i) = 0.8 * sUn1(j, i) + 0.05 * sUn1(j, (i + 1) % N) + 0.05 * sUn1(j, (i - 1 + N) % N)
|
||||
+ 0.05 * sUn1((j + 1) % N, i) + 0.05 * sUn1((j - 1 + N) % N, i);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
stream_ctx ctx;
|
||||
|
||||
size_t NITER = 500;
|
||||
size_t N = 1000;
|
||||
bool vtk_dump = false;
|
||||
|
||||
if (argc > 1)
|
||||
{
|
||||
NITER = atoi(argv[1]);
|
||||
}
|
||||
|
||||
if (argc > 2)
|
||||
{
|
||||
N = atoi(argv[2]);
|
||||
}
|
||||
|
||||
if (argc > 3)
|
||||
{
|
||||
int val = atoi(argv[3]);
|
||||
vtk_dump = (val == 1);
|
||||
}
|
||||
|
||||
size_t TOTAL_SIZE = N * N;
|
||||
|
||||
double* Un = new double[TOTAL_SIZE];
|
||||
double* Un1 = new double[TOTAL_SIZE];
|
||||
|
||||
for (size_t idx = 0; idx < TOTAL_SIZE; idx++)
|
||||
{
|
||||
Un[idx] = (idx == 0) ? 1.0 : 0.0;
|
||||
Un1[idx] = Un[idx];
|
||||
}
|
||||
|
||||
auto lUn = ctx.logical_data(make_slice(Un, std::tuple{N, N}, N));
|
||||
auto lUn1 = ctx.logical_data(make_slice(Un1, std::tuple{N, N}, N));
|
||||
|
||||
// std::shared_ptr<execution_grid> all_devs = exec_place::all_devices();
|
||||
// use grid [ 0 0 0 0 ] for debugging purpose
|
||||
auto all_devs = exec_place::repeat(exec_place::device(0), 4);
|
||||
|
||||
// Partition over the vector of processor along the y-axis of the data domain
|
||||
// TODO implement the proper tiled_partitioning along y !
|
||||
data_place cdp = data_place::composite(tiled_partition<128>(), all_devs);
|
||||
|
||||
for (size_t iter = 0; iter < NITER; iter++)
|
||||
{
|
||||
// UPDATE Un from Un1
|
||||
ctx.task(lUn.rw(cdp), lUn1.read(cdp))->*[&](auto stream, auto sUn, auto sUn1) {
|
||||
stencil2D_kernel<double><<<32, 128, 0, stream>>>(sUn, sUn1);
|
||||
};
|
||||
|
||||
// We make sure that the total sum of elements remains constant
|
||||
if (iter % 250 == 0)
|
||||
{
|
||||
double sum = 0.0;
|
||||
|
||||
ctx.task(exec_place::host(), lUn.read())->*[&](auto stream, auto sUn) {
|
||||
cuda_safe_call(cudaStreamSynchronize(stream));
|
||||
for (size_t j = 0; j < N; j++)
|
||||
{
|
||||
for (size_t i = 0; i < N; i++)
|
||||
{
|
||||
sum += sUn(j, i);
|
||||
}
|
||||
}
|
||||
|
||||
if (vtk_dump)
|
||||
{
|
||||
char str[32];
|
||||
snprintf(str, 32, "Un_%05zu.vtk", iter);
|
||||
mdspan_to_vtk(sUn, std::string(str));
|
||||
}
|
||||
};
|
||||
|
||||
// fprintf(stderr, "iter %d : CHECK SUM = %e\n", iter, sum);
|
||||
}
|
||||
|
||||
std::swap(lUn, lUn1);
|
||||
}
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,750 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
/**
|
||||
* @file
|
||||
*
|
||||
* @brief This example implements a Cholesky decomposition over multiple devices using CUBLAS and CUSOLVER
|
||||
*
|
||||
* It also illustrates how we can use CUDASTF to allocate temporary data for CUSOLVER in CUDASTF tasks
|
||||
*/
|
||||
|
||||
#include <cuda/experimental/__stf/stream/interfaces/slice_reduction_ops.cuh>
|
||||
#include <cuda/experimental/__stf/utility/nvtx.cuh>
|
||||
#include <cuda/experimental/stf.cuh>
|
||||
|
||||
#include <iostream>
|
||||
|
||||
#define TILED
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
// Global for the sake of simplicity !
|
||||
stream_ctx ctx;
|
||||
|
||||
/* Get a CUBLAS handle valid on the current execution place, or initialize it lazily */
|
||||
cublasHandle_t& get_cublas_handle(const exec_place& ep = exec_place::current_device())
|
||||
{
|
||||
static std::unordered_map<exec_place, cublasHandle_t, hash<exec_place>> cublas_handles;
|
||||
auto& result = cublas_handles[ep];
|
||||
if (result == cublasHandle_t())
|
||||
{ // not found, default value inserted
|
||||
// Lazy initialization, and save the handle for future use
|
||||
cuda_safe_call(cublasCreate(&result));
|
||||
}
|
||||
return result;
|
||||
}
|
||||
|
||||
/* Get a CUSOLVER handle valid on the current execution place, or initialize it lazily */
|
||||
cusolverDnHandle_t& get_cusolver_handle(const exec_place& ep = exec_place::current_device())
|
||||
{
|
||||
static std::unordered_map<exec_place, cusolverDnHandle_t, hash<exec_place>> cusolver_handles;
|
||||
auto& result = cusolver_handles[ep];
|
||||
if (result == cusolverDnHandle_t())
|
||||
{ // not found, default value inserted
|
||||
// Lazy initialization, and save the handle for future use
|
||||
cuda_safe_call(cusolverDnCreate(&result));
|
||||
}
|
||||
return result;
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
class matrix
|
||||
{
|
||||
public:
|
||||
matrix(int NROWS, int NCOLS, int BLOCKSIZE_ROWS, int BLOCKSIZE_COLS, bool is_sym, const char* _symbol = "matrix")
|
||||
{
|
||||
symbol = _symbol;
|
||||
|
||||
sym_matrix = is_sym;
|
||||
|
||||
m = NROWS;
|
||||
mb = BLOCKSIZE_ROWS;
|
||||
|
||||
n = NCOLS;
|
||||
nb = BLOCKSIZE_COLS;
|
||||
|
||||
assert(m % mb == 0);
|
||||
assert(n % nb == 0);
|
||||
|
||||
// cuda_safe_call(cudaMallocHost(&h_array, m*n*sizeof(T)));
|
||||
// fprintf(stderr, "Allocating %ld x %ld x %ld = %ld bytes (%f GB) on host for %s\n", m, n, sizeof(T), s,
|
||||
// s / (1024.0 * 1024.0 * 1024.0), _symbol);
|
||||
h_array.resize(m * n);
|
||||
cuda_safe_call(cudaHostRegister(&h_array[0], h_array.size() * sizeof(T), cudaHostRegisterPortable));
|
||||
|
||||
// Compute the number of blocks
|
||||
mt = m / mb;
|
||||
nt = n / nb;
|
||||
|
||||
handles.resize(mt * nt);
|
||||
|
||||
for (size_t colb = 0; colb < nt; colb++)
|
||||
{
|
||||
size_t low_rowb = sym_matrix ? colb : 0;
|
||||
for (size_t rowb = low_rowb; rowb < mt; rowb++)
|
||||
{
|
||||
T* addr_h = get_block_h(rowb, colb);
|
||||
auto& h = handle(rowb, colb);
|
||||
|
||||
#ifdef TILED
|
||||
// tiles are stored contiguously
|
||||
size_t ld = mb;
|
||||
#else
|
||||
size_t ld = m;
|
||||
#endif
|
||||
std::ignore = ld; // work around bug in compiler
|
||||
h = ctx.logical_data(make_slice(addr_h, std::tuple{mb, nb}, ld));
|
||||
h.set_symbol(std::string(symbol) + "_" + std::to_string(rowb) + "_" + std::to_string(colb));
|
||||
h.set_write_back(false);
|
||||
}
|
||||
}
|
||||
|
||||
cuda_safe_call(cudaGetDeviceCount(&ndevs));
|
||||
for (int a = 1; a * a <= ndevs; a++)
|
||||
{
|
||||
if (ndevs % a == 0)
|
||||
{
|
||||
grid_p = a;
|
||||
grid_q = ndevs / a;
|
||||
}
|
||||
}
|
||||
|
||||
assert(grid_p * grid_q == ndevs);
|
||||
|
||||
// std::cout << "FOUND " << ndevs << " DEVICES "
|
||||
// << "p=" << grid_p << " q=" << grid_q << '\n';
|
||||
}
|
||||
|
||||
int get_preferred_devid(int row, int col)
|
||||
{
|
||||
return (row % grid_p) + (col % grid_q) * grid_p;
|
||||
}
|
||||
|
||||
auto& handle(int row, int col)
|
||||
{
|
||||
return handles[row + col * mt];
|
||||
}
|
||||
|
||||
size_t get_index(size_t row, size_t col)
|
||||
{
|
||||
#ifdef TILED
|
||||
// Find which tile contains this element
|
||||
int tile_row = row / mb;
|
||||
int tile_col = col / nb;
|
||||
|
||||
size_t tile_size = mb * nb;
|
||||
|
||||
// Look for the index of the beginning of the tile
|
||||
size_t tile_start = (tile_row + mt * tile_col) * tile_size;
|
||||
|
||||
// Offset within the tile
|
||||
size_t offset = (row % mb) + (col % nb) * mb;
|
||||
|
||||
return tile_start + offset;
|
||||
#else
|
||||
return row + col * m;
|
||||
#endif
|
||||
}
|
||||
|
||||
T* get_block_h(int brow, int bcol)
|
||||
{
|
||||
size_t index = get_index(brow * mb, bcol * nb);
|
||||
return &h_array[index];
|
||||
}
|
||||
|
||||
// Fill with func(Matrix*,row, col)
|
||||
template <typename Fun>
|
||||
void fill(Fun&& fun)
|
||||
{
|
||||
nvtx_range r("fill");
|
||||
// Fill blocks by blocks
|
||||
for (size_t colb = 0; colb < nt; colb++)
|
||||
{
|
||||
size_t low_rowb = sym_matrix ? colb : 0;
|
||||
for (size_t rowb = low_rowb; rowb < mt; rowb++)
|
||||
{
|
||||
// Each task fills a block
|
||||
auto& h = handle(rowb, colb);
|
||||
int devid = get_preferred_devid(rowb, colb);
|
||||
|
||||
ctx.parallel_for(exec_place::device(devid), h.shape(), h.write()).set_symbol("INIT")->*
|
||||
[=] __device__(size_t lrow, size_t lcol, auto sA) {
|
||||
size_t row = lrow + rowb * sA.extent(0);
|
||||
size_t col = lcol + colb * sA.extent(1);
|
||||
sA(lrow, lcol) = fun(row, col);
|
||||
};
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
std::vector<T> h_array;
|
||||
size_t m; // nrows
|
||||
size_t n; // ncols
|
||||
|
||||
// Is this a sym matrix ? (lower assumed)
|
||||
bool sym_matrix;
|
||||
|
||||
size_t mb; // block size (rows)
|
||||
size_t nb; // block size (cols)
|
||||
|
||||
size_t mt; // number of column blocks
|
||||
size_t nt; // number of row blocks
|
||||
|
||||
// abstract data handles
|
||||
std::vector<logical_data<slice<double, 2>>> handles;
|
||||
|
||||
const char* symbol;
|
||||
|
||||
// for the mapping
|
||||
int ndevs;
|
||||
int grid_p, grid_q;
|
||||
};
|
||||
|
||||
void DPOTRF(cublasFillMode_t uplo, class matrix<double>& A, int A_row, int A_col)
|
||||
{
|
||||
auto& Akk = A.handle(A_row, A_col);
|
||||
size_t m_akk = Akk.shape().extent(0);
|
||||
// Note that the handle may be different from the actual handle...
|
||||
int Lwork_expected;
|
||||
cuda_safe_call(cusolverDnDpotrf_bufferSize(get_cusolver_handle(), uplo, m_akk, nullptr, 0, &Lwork_expected));
|
||||
|
||||
auto potrf_buffer = ctx.logical_data<double>(Lwork_expected);
|
||||
auto devInfo = ctx.logical_data(shape_of<slice<int>>(1));
|
||||
|
||||
auto t =
|
||||
ctx.task(exec_place::device(A.get_preferred_devid(A_row, A_col)), Akk.rw(), potrf_buffer.write(), devInfo.write());
|
||||
t.set_symbol("DPOTRF");
|
||||
t->*[uplo](cudaStream_t s, auto sAkk, auto buffer, auto info) {
|
||||
auto& h = get_cusolver_handle();
|
||||
cuda_safe_call(cusolverDnSetStream(h, s));
|
||||
|
||||
cuda_safe_call(cusolverDnDpotrf(
|
||||
h,
|
||||
uplo,
|
||||
sAkk.extent(0),
|
||||
sAkk.data_handle(),
|
||||
sAkk.stride(1),
|
||||
buffer.data_handle(),
|
||||
buffer.extent(0),
|
||||
info.data_handle()));
|
||||
};
|
||||
}
|
||||
|
||||
void DGEMM(
|
||||
cublasOperation_t transa,
|
||||
cublasOperation_t transb,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
int A_row,
|
||||
int A_col,
|
||||
class matrix<double>& B,
|
||||
int B_row,
|
||||
int B_col,
|
||||
double beta,
|
||||
class matrix<double>& C,
|
||||
int C_row,
|
||||
int C_col)
|
||||
{
|
||||
auto redux_op = std::make_shared<slice_reduction_op_sum<double, 2>>();
|
||||
|
||||
// If beta == 1.0 (we assume this is exactly 1.0), then this operation is
|
||||
// an accumulation with the add operator
|
||||
auto dep_c = (beta == 1.0) ? C.handle(C_row, C_col).relaxed(redux_op) : C.handle(C_row, C_col).rw();
|
||||
auto t = ctx.task(exec_place::device(A.get_preferred_devid(C_row, C_col)),
|
||||
A.handle(A_row, A_col).read(),
|
||||
B.handle(B_row, B_col).read(),
|
||||
dep_c);
|
||||
t.set_symbol("DGEMM");
|
||||
t->*[transa, transb, alpha, beta](cudaStream_t s, auto sA, auto sB, auto sC) {
|
||||
EXPECT(sC.data_handle() != nullptr);
|
||||
auto& h = get_cublas_handle();
|
||||
cuda_safe_call(cublasSetStream(h, s));
|
||||
|
||||
auto k = (transa == CUBLAS_OP_N) ? sA.extent(1) : sA.extent(0);
|
||||
cuda_safe_call(cublasDgemm(
|
||||
h,
|
||||
transa,
|
||||
transb,
|
||||
sC.extent(0),
|
||||
sC.extent(1),
|
||||
k,
|
||||
&alpha,
|
||||
sA.data_handle(),
|
||||
sA.stride(1),
|
||||
sB.data_handle(),
|
||||
sB.stride(1),
|
||||
&beta,
|
||||
sC.data_handle(),
|
||||
sC.stride(1)));
|
||||
};
|
||||
}
|
||||
|
||||
void DSYRK(
|
||||
cublasFillMode_t uplo,
|
||||
cublasOperation_t trans,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
int A_row,
|
||||
int A_col,
|
||||
double beta,
|
||||
class matrix<double>& C,
|
||||
int C_row,
|
||||
int C_col)
|
||||
{
|
||||
auto t = ctx.task(exec_place::device(A.get_preferred_devid(C_row, C_col)),
|
||||
A.handle(A_row, A_col).read(),
|
||||
C.handle(C_row, C_col).rw());
|
||||
t.set_symbol("DSYRK");
|
||||
t->*[uplo, trans, alpha, beta](cudaStream_t s, auto sA, auto sC) {
|
||||
auto& h = get_cublas_handle();
|
||||
cuda_safe_call(cublasSetStream(h, s));
|
||||
|
||||
// number of rows of matrix op(A) and C
|
||||
auto n = sC.extent(0);
|
||||
|
||||
// number of columns of matrix op(A)
|
||||
auto k = (trans == CUBLAS_OP_N) ? sA.extent(1) : sA.extent(0);
|
||||
|
||||
cuda_safe_call(
|
||||
cublasDsyrk(h, uplo, trans, n, k, &alpha, sA.data_handle(), sA.stride(1), &beta, sC.data_handle(), sC.stride(1)));
|
||||
};
|
||||
}
|
||||
|
||||
void DTRSM(
|
||||
cublasSideMode_t side,
|
||||
cublasFillMode_t uplo,
|
||||
cublasOperation_t transa,
|
||||
cublasDiagType_t diag,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
int A_row,
|
||||
int A_col,
|
||||
class matrix<double>& B,
|
||||
int B_row,
|
||||
int B_col)
|
||||
{
|
||||
auto t = ctx.task(exec_place::device(A.get_preferred_devid(B_row, B_col)),
|
||||
A.handle(A_row, A_col).read(),
|
||||
B.handle(B_row, B_col).rw());
|
||||
t.set_symbol("DTRSM");
|
||||
t->*[side, uplo, transa, diag, alpha](cudaStream_t s, auto sA, auto sB) {
|
||||
auto& h = get_cublas_handle();
|
||||
cuda_safe_call(cublasSetStream(h, s));
|
||||
|
||||
cuda_safe_call(cublasDtrsm(
|
||||
h,
|
||||
side,
|
||||
uplo,
|
||||
transa,
|
||||
diag,
|
||||
sB.extent(0),
|
||||
sB.extent(1),
|
||||
&alpha,
|
||||
sA.data_handle(),
|
||||
sA.stride(1),
|
||||
sB.data_handle(),
|
||||
sB.stride(1)));
|
||||
};
|
||||
}
|
||||
|
||||
void PDNRM2_HOST(matrix<double>* A, double* result)
|
||||
{
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("red");
|
||||
#endif
|
||||
|
||||
for (size_t rowb = 0; rowb < A->mt; rowb++)
|
||||
{
|
||||
for (size_t colb = 0; colb < A->nt; colb++)
|
||||
{
|
||||
ctx.host_launch(A->handle(rowb, colb).read())->*[=](auto sA) {
|
||||
double res2 = 0.0;
|
||||
for (size_t col = 0; col < sA.extent(1); col++)
|
||||
{
|
||||
for (size_t row = 0; row < sA.extent(0); row++)
|
||||
{
|
||||
double v = sA(row, col);
|
||||
res2 += v * v;
|
||||
}
|
||||
}
|
||||
*result += res2;
|
||||
};
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
void PDPOTRF(matrix<double>& A)
|
||||
{
|
||||
nvtx_range r("PDPOTRF");
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("yellow");
|
||||
#endif
|
||||
|
||||
assert(A.m == A.n);
|
||||
assert(A.mt == A.nt);
|
||||
|
||||
int NBLOCKS = A.mt;
|
||||
assert(A.mb == A.nb);
|
||||
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
|
||||
for (int K = 0; K < NBLOCKS; K++)
|
||||
{
|
||||
int dev_akk = A.get_preferred_devid(K, K);
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(K, K)));
|
||||
DPOTRF(CUBLAS_FILL_MODE_LOWER, A, K, K);
|
||||
|
||||
for (int row = K + 1; row < NBLOCKS; row++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, K)));
|
||||
DTRSM(CUBLAS_SIDE_RIGHT, CUBLAS_FILL_MODE_LOWER, CUBLAS_OP_T, CUBLAS_DIAG_NON_UNIT, 1.0, A, K, K, A, row, K);
|
||||
|
||||
for (int col = K + 1; col < row; col++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, col)));
|
||||
DGEMM(CUBLAS_OP_N, CUBLAS_OP_T, -1.0, A, row, K, A, col, K, 1.0, A, row, col);
|
||||
}
|
||||
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, row)));
|
||||
DSYRK(CUBLAS_FILL_MODE_LOWER, CUBLAS_OP_N, -1.0, A, row, K, 1.0, A, row, row);
|
||||
}
|
||||
}
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
}
|
||||
|
||||
// Algorithm from PLASMA
|
||||
void PDTRSM(cublasSideMode_t side,
|
||||
cublasFillMode_t uplo,
|
||||
cublasOperation_t trans,
|
||||
cublasDiagType_t diag,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
class matrix<double>& B)
|
||||
{
|
||||
nvtx_range r("PDTRSM");
|
||||
|
||||
// std::cout << "[PDTRSM] START B MT " << B.mt << " NT " << B.nt << '\n';
|
||||
|
||||
if (side == CUBLAS_SIDE_LEFT)
|
||||
{
|
||||
if (uplo == CUBLAS_FILL_MODE_UPPER)
|
||||
{
|
||||
// TODO
|
||||
assert(0);
|
||||
abort();
|
||||
}
|
||||
else
|
||||
{
|
||||
//===========================================
|
||||
// CUBLAS_SIDE_LEFT / CUBLAS_FILL_MODE_LOWER / CUBLAS_OP_N
|
||||
//===========================================
|
||||
if (trans == CUBLAS_OP_N)
|
||||
{
|
||||
for (size_t k = 0; k < B.mt; k++)
|
||||
{
|
||||
double lalpha = k == 0 ? alpha : 1.0;
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(k, k)));
|
||||
DTRSM(side, uplo, trans, diag, lalpha, A, k, k, B, k, n);
|
||||
}
|
||||
for (size_t m = k + 1; m < B.mt; m++)
|
||||
{
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(m, k)));
|
||||
DGEMM(CUBLAS_OP_N, CUBLAS_OP_N, -1.0, A, m, k, B, k, n, lalpha, B, m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
//================================================
|
||||
// CUBLAS_SIDE_LEFT / CUBLAS_FILL_MODE_LOWER / CUBLAS_OP_[C|T]
|
||||
//================================================
|
||||
else
|
||||
{
|
||||
for (size_t k = 0; k < B.mt; k++)
|
||||
{
|
||||
double lalpha = k == 0 ? alpha : 1.0;
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(B.mt - k - 1, B.mt - k - 1)));
|
||||
DTRSM(side, uplo, trans, diag, lalpha, A, B.mt - k - 1, B.mt - k - 1, B, B.mt - k - 1, n);
|
||||
}
|
||||
for (size_t m = k + 1; m < B.mt; m++)
|
||||
{
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(B.mt - k - 1, B.mt - 1 - m)));
|
||||
DGEMM(
|
||||
trans, CUBLAS_OP_N, -1.0, A, B.mt - k - 1, B.mt - 1 - m, B, B.mt - k - 1, n, lalpha, B, B.mt - 1 - m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
else
|
||||
{
|
||||
// TODO
|
||||
abort();
|
||||
}
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
// std::cout << "[PDTRSM] END" << '\n';
|
||||
}
|
||||
|
||||
void PDPOTRS(matrix<double>& A, class matrix<double>& B, cublasFillMode_t uplo)
|
||||
{
|
||||
nvtx_range r("PDPOTRS");
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("green");
|
||||
#endif
|
||||
|
||||
// std::cout << "[PDPOTRS] START" << '\n';
|
||||
// Call the parallel functions.
|
||||
PDTRSM(
|
||||
CUBLAS_SIDE_LEFT, uplo, uplo == CUBLAS_FILL_MODE_UPPER ? CUBLAS_OP_T : CUBLAS_OP_N, CUBLAS_DIAG_NON_UNIT, 1.0, A, B);
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("darkgreen");
|
||||
#endif
|
||||
|
||||
PDTRSM(
|
||||
CUBLAS_SIDE_LEFT, uplo, uplo == CUBLAS_FILL_MODE_UPPER ? CUBLAS_OP_N : CUBLAS_OP_T, CUBLAS_DIAG_NON_UNIT, 1.0, A, B);
|
||||
// std::cout << "[PDPOTRS] END" << '\n';
|
||||
}
|
||||
|
||||
/*****************************************************************************
|
||||
* Parallel tile matrix-matrix
|
||||
*multiplication.
|
||||
* @see plasma_omp_dgemm
|
||||
******************************************************************************/
|
||||
void PDGEMM(cublasOperation_t transa,
|
||||
cublasOperation_t transb,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
class matrix<double>& B,
|
||||
double beta,
|
||||
class matrix<double>& C)
|
||||
{
|
||||
nvtx_range r("PDGEMM");
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("blue");
|
||||
#endif
|
||||
|
||||
for (size_t m = 0; m < C.mt; m++)
|
||||
{
|
||||
for (size_t n = 0; n < C.nt; n++)
|
||||
{
|
||||
//=========================================
|
||||
// alpha*A*B does not contribute; scale C
|
||||
//=========================================
|
||||
int inner_k = transa == CUBLAS_OP_N ? A.n : A.m;
|
||||
if (alpha == 0.0 || inner_k == 0)
|
||||
{
|
||||
DGEMM(transa, transb, alpha, A, 0, 0, B, 0, 0, beta, C, m, n);
|
||||
}
|
||||
else if (transa == CUBLAS_OP_N)
|
||||
{
|
||||
//================================
|
||||
// CUBLAS_OP_N / CUBLAS_OP_N
|
||||
//================================
|
||||
if (transb == CUBLAS_OP_N)
|
||||
{
|
||||
for (size_t k = 0; k < A.nt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, m, k, B, k, n, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
//=====================================
|
||||
// CUBLAS_OP_N / CUBLAS_OP_T
|
||||
//=====================================
|
||||
else
|
||||
{
|
||||
for (size_t k = 0; k < A.nt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, m, k, B, n, k, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
else
|
||||
{
|
||||
//=====================================
|
||||
// CUBLAS_OP_T / CUBLAS_OP_N
|
||||
//=====================================
|
||||
if (transb == CUBLAS_OP_N)
|
||||
{
|
||||
for (size_t k = 0; k < A.mt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, k, m, B, k, n, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
//==========================================
|
||||
// CUBLAS_OP_T / CUBLAS_OP_T
|
||||
//==========================================
|
||||
else
|
||||
{
|
||||
for (size_t k = 0; k < A.mt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, k, m, B, n, k, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
int N = 1024;
|
||||
int NB = 128;
|
||||
|
||||
if (argc > 1)
|
||||
{
|
||||
N = atoi(argv[1]);
|
||||
}
|
||||
|
||||
if (argc > 2)
|
||||
{
|
||||
NB = atoi(argv[2]);
|
||||
}
|
||||
|
||||
int check_result = 1;
|
||||
if (getenv("CHECK_RESULT"))
|
||||
{
|
||||
check_result = atoi(getenv("CHECK_RESULT"));
|
||||
}
|
||||
|
||||
assert(N % NB == 0);
|
||||
|
||||
// Use pools of preallocated blocks
|
||||
auto fixed_alloc = block_allocator<fixed_size_allocator>(ctx, NB * NB * sizeof(double));
|
||||
ctx.set_allocator(fixed_alloc);
|
||||
|
||||
// Set up CUBLAS and CUSOLVER
|
||||
int ndevs;
|
||||
cuda_safe_call(cudaGetDeviceCount(&ndevs));
|
||||
|
||||
for (int d = 0; d < ndevs; d++)
|
||||
{
|
||||
auto lX = ctx.logical_data(shape_of<slice<double>>(1));
|
||||
ctx.parallel_for(exec_place::device(d), lX.shape(), lX.write())->*[] __device__(size_t, auto) {};
|
||||
cuda_safe_call(cudaSetDevice(d));
|
||||
get_cublas_handle();
|
||||
get_cusolver_handle();
|
||||
}
|
||||
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
|
||||
matrix<double> A(N, N, NB, NB, true, "A");
|
||||
matrix<double> Aref(N, N, NB, NB, false, "Aref");
|
||||
|
||||
// (Hilbert matrix + 2*N*Id) to have a diagonal dominant matrix
|
||||
auto hilbert = [=] __host__ __device__(size_t row, size_t col) {
|
||||
return 1.0 / (col + row + 1.0) + 2.0 * N * (col == row);
|
||||
};
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
Aref.fill(hilbert);
|
||||
}
|
||||
|
||||
A.fill(hilbert);
|
||||
|
||||
/* Right-hand side */
|
||||
matrix<double> B_potrs(N, 1, NB, 1, false, "B");
|
||||
matrix<double> Bref_potrs(N, 1, NB, 1, false, "Bref");
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
auto rhs_vals = [] __host__ __device__(size_t row, size_t /*unused*/) { return 1.0 * (row + 1); };
|
||||
B_potrs.fill(rhs_vals);
|
||||
Bref_potrs.fill(rhs_vals);
|
||||
}
|
||||
|
||||
// // Compute ||Bref||
|
||||
double Bref_nrm2 = 0.0;
|
||||
double res_nrm2 = 0.0;
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
PDNRM2_HOST(&Bref_potrs, &Bref_nrm2);
|
||||
}
|
||||
|
||||
cudaEvent_t startEvent_pdpotrf, stopEvent_pdpotrf;
|
||||
float milliseconds_pdpotrf = 0;
|
||||
|
||||
// for (size_t row = 0; row < A.mt; row++)
|
||||
// {
|
||||
// for (size_t col = 0; col <= row; col++)
|
||||
// {
|
||||
// cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, col)));
|
||||
// NOOP(A, row, col);
|
||||
// }
|
||||
// }
|
||||
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
|
||||
cuda_safe_call(cudaStreamSynchronize(ctx.fence()));
|
||||
|
||||
cuda_safe_call(cudaEventCreate(&startEvent_pdpotrf));
|
||||
cuda_safe_call(cudaEventCreate(&stopEvent_pdpotrf));
|
||||
|
||||
cuda_safe_call(cudaEventRecord(startEvent_pdpotrf, ctx.fence()));
|
||||
|
||||
PDPOTRF(A);
|
||||
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
cuda_safe_call(cudaEventRecord(stopEvent_pdpotrf, ctx.fence()));
|
||||
|
||||
/*
|
||||
* POTRS
|
||||
*/
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
// Solve AX = B and put the result in B
|
||||
PDPOTRS(A, B_potrs, CUBLAS_FILL_MODE_LOWER);
|
||||
|
||||
// Compute (AX - B)
|
||||
// Bref = (Aref*B - Bref)
|
||||
PDGEMM(CUBLAS_OP_N, CUBLAS_OP_N, 1.0, Aref, B_potrs, -1.0, Bref_potrs);
|
||||
|
||||
// Compute ||AX - B|| = ||Bref||
|
||||
PDNRM2_HOST(&Bref_potrs, &res_nrm2);
|
||||
}
|
||||
|
||||
ctx.finalize();
|
||||
|
||||
cuda_safe_call(cudaEventElapsedTime(&milliseconds_pdpotrf, startEvent_pdpotrf, stopEvent_pdpotrf));
|
||||
|
||||
double gflops_pdpotrf = 1.0 / 3.0 * ((double) N * (double) N * (double) N) / (1000000000.0);
|
||||
std::cout << "[PDPOTRF] ELAPSED: " << milliseconds_pdpotrf
|
||||
<< " ms, GFLOPS: " << gflops_pdpotrf / (milliseconds_pdpotrf / 1000.0) << '\n';
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
if (double residual = sqrt(res_nrm2) / sqrt(Bref_nrm2); residual >= 0.01)
|
||||
{
|
||||
std::cerr << "[POTRS] ||AX - B|| : " << sqrt(res_nrm2) << '\n';
|
||||
std::cerr << "[POTRS] ||B|| : " << sqrt(Bref_nrm2) << '\n';
|
||||
std::cerr << "[POTRS] RESIDUAL (||AX - B||/||B||) : " << residual << '\n';
|
||||
assert(!"Algorithm did not converge.");
|
||||
}
|
||||
}
|
||||
}
|
||||
@@ -1,693 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/__stf/stream/stream_ctx.cuh>
|
||||
#include <cuda/experimental/__stf/utility/nvtx.cuh>
|
||||
|
||||
#define TILED
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
// The backend used in this example only depends on that type
|
||||
using backend_type = stream_ctx;
|
||||
// using backend_type = graph_ctx;
|
||||
|
||||
// Global for the sake of simplicity !
|
||||
backend_type ctx;
|
||||
|
||||
/* Get a CUBLAS handle valid on the current execution place, or initialize it lazily */
|
||||
cublasHandle_t& get_cublas_handle(const exec_place& ep = exec_place::current_device())
|
||||
{
|
||||
static std::unordered_map<exec_place, cublasHandle_t, hash<exec_place>> cublas_handles;
|
||||
auto& result = cublas_handles[ep];
|
||||
if (result == cublasHandle_t())
|
||||
{ // not found, default value inserted
|
||||
// Lazy initialization, and save the handle for future use
|
||||
cuda_safe_call(cublasCreate(&result));
|
||||
}
|
||||
return result;
|
||||
}
|
||||
|
||||
/* Get a CUSOLVER handle valid on the current execution place, or initialize it lazily */
|
||||
cusolverDnHandle_t& get_cusolver_handle(const exec_place& ep = exec_place::current_device())
|
||||
{
|
||||
static std::unordered_map<exec_place, cusolverDnHandle_t, hash<exec_place>> cusolver_handles;
|
||||
auto& result = cusolver_handles[ep];
|
||||
if (result == cusolverDnHandle_t())
|
||||
{ // not found, default value inserted
|
||||
// Lazy initialization, and save the handle for future use
|
||||
cuda_safe_call(cusolverDnCreate(&result));
|
||||
}
|
||||
return result;
|
||||
}
|
||||
|
||||
template <typename T>
|
||||
class matrix
|
||||
{
|
||||
public:
|
||||
matrix(int NROWS, int NCOLS, int BLOCKSIZE_ROWS, int BLOCKSIZE_COLS, bool is_sym, const char* _symbol = "matrix")
|
||||
{
|
||||
symbol = _symbol;
|
||||
|
||||
sym_matrix = is_sym;
|
||||
|
||||
m = NROWS;
|
||||
mb = BLOCKSIZE_ROWS;
|
||||
|
||||
n = NCOLS;
|
||||
nb = BLOCKSIZE_COLS;
|
||||
|
||||
assert(m % mb == 0);
|
||||
assert(n % nb == 0);
|
||||
|
||||
// cuda_safe_call(cudaMallocHost(&h_array, m*n*sizeof(T)));
|
||||
// fprintf(stderr, "Allocating %ld x %ld x %ld = %ld bytes (%f GB) on host for %s\n", m, n, sizeof(T), s,
|
||||
// s / (1024.0 * 1024.0 * 1024.0), _symbol);
|
||||
h_array.resize(m * n);
|
||||
cuda_safe_call(cudaHostRegister(&h_array[0], h_array.size() * sizeof(T), cudaHostRegisterPortable));
|
||||
|
||||
// Compute the number of blocks
|
||||
mt = m / mb;
|
||||
nt = n / nb;
|
||||
|
||||
handles.resize(mt * nt);
|
||||
|
||||
for (size_t colb = 0; colb < nt; colb++)
|
||||
{
|
||||
int low_rowb = sym_matrix ? colb : 0;
|
||||
for (size_t rowb = low_rowb; rowb < mt; rowb++)
|
||||
{
|
||||
T* addr_h = get_block_h(rowb, colb);
|
||||
auto& h = handle(rowb, colb);
|
||||
|
||||
#ifdef TILED
|
||||
// tiles are stored contiguously
|
||||
size_t ld = mb;
|
||||
#else
|
||||
size_t ld = m;
|
||||
#endif
|
||||
std::ignore = ld; // work around bug in compiler
|
||||
h = ctx.logical_data(make_slice(addr_h, std::tuple{mb, nb}, ld));
|
||||
h.set_symbol(std::string(symbol) + "_" + std::to_string(rowb) + "_" + std::to_string(colb));
|
||||
}
|
||||
}
|
||||
|
||||
cuda_safe_call(cudaGetDeviceCount(&ndevs));
|
||||
for (int a = 1; a * a <= ndevs; a++)
|
||||
{
|
||||
if (ndevs % a == 0)
|
||||
{
|
||||
grid_p = a;
|
||||
grid_q = ndevs / a;
|
||||
}
|
||||
}
|
||||
|
||||
assert(grid_p * grid_q == ndevs);
|
||||
|
||||
// std::cout << "FOUND " << ndevs << " DEVICES "
|
||||
// << "p=" << grid_p << " q=" << grid_q << '\n';
|
||||
}
|
||||
|
||||
int get_preferred_devid(int row, int col)
|
||||
{
|
||||
return (row % grid_p) + (col % grid_q) * grid_p;
|
||||
}
|
||||
|
||||
auto& handle(int row, int col)
|
||||
{
|
||||
return handles[row + col * mt];
|
||||
}
|
||||
|
||||
size_t get_index(size_t row, size_t col)
|
||||
{
|
||||
#ifdef TILED
|
||||
// Find which tile contains this element
|
||||
int tile_row = row / mb;
|
||||
int tile_col = col / nb;
|
||||
|
||||
size_t tile_size = mb * nb;
|
||||
|
||||
// Look for the index of the beginning of the tile
|
||||
size_t tile_start = (tile_row + mt * tile_col) * tile_size;
|
||||
|
||||
// Offset within the tile
|
||||
size_t offset = (row % mb) + (col % nb) * mb;
|
||||
|
||||
return tile_start + offset;
|
||||
#else
|
||||
return row + col * m;
|
||||
#endif
|
||||
}
|
||||
|
||||
T* get_block_h(int brow, int bcol)
|
||||
{
|
||||
size_t index = get_index(brow * mb, bcol * nb);
|
||||
return &h_array[index];
|
||||
}
|
||||
|
||||
// Fill with func(Matrix*,row, col)
|
||||
template <typename Fun>
|
||||
void fill(Fun&& fun)
|
||||
{
|
||||
// Fill blocks by blocks
|
||||
for (size_t colb = 0; colb < nt; colb++)
|
||||
{
|
||||
size_t low_rowb = sym_matrix ? colb : 0;
|
||||
for (size_t rowb = low_rowb; rowb < mt; rowb++)
|
||||
{
|
||||
// Each task fills a block
|
||||
ctx.host_launch(handle(rowb, colb).write())->*[=, self = this](auto sA) {
|
||||
for (size_t lcol = 0; lcol < sA.extent(1); lcol++)
|
||||
{
|
||||
size_t col = lcol + colb * sA.extent(1);
|
||||
for (size_t lrow = 0; lrow < sA.extent(0); lrow++)
|
||||
{
|
||||
size_t row = lrow + rowb * sA.extent(0);
|
||||
sA(lrow, lcol) = fun(*self, row, col);
|
||||
}
|
||||
}
|
||||
};
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
std::vector<T> h_array;
|
||||
size_t m; // nrows
|
||||
size_t n; // ncols
|
||||
|
||||
// Is this a sym matrix ? (lower assumed)
|
||||
bool sym_matrix;
|
||||
|
||||
size_t mb; // block size (rows)
|
||||
size_t nb; // block size (cols)
|
||||
|
||||
size_t mt; // number of column blocks
|
||||
size_t nt; // number of row blocks
|
||||
|
||||
// abstract data handles
|
||||
std::vector<logical_data<slice<double, 2>>> handles;
|
||||
|
||||
const char* symbol;
|
||||
|
||||
// for the mapping
|
||||
int ndevs;
|
||||
int grid_p, grid_q;
|
||||
};
|
||||
|
||||
void DPOTRF(cublasFillMode_t uplo, class matrix<double>& A, int A_row, int A_col)
|
||||
{
|
||||
auto& Akk = A.handle(A_row, A_col);
|
||||
size_t m_akk = Akk.shape().extent(0);
|
||||
// Note that the handle may be different from the actual handle...
|
||||
int Lwork_expected;
|
||||
cuda_safe_call(cusolverDnDpotrf_bufferSize(get_cusolver_handle(), uplo, m_akk, nullptr, 0, &Lwork_expected));
|
||||
|
||||
auto potrf_buffer = ctx.logical_data<double>(Lwork_expected);
|
||||
auto devInfo = ctx.logical_data(shape_of<slice<int>>(1));
|
||||
|
||||
auto t = ctx.task(Akk.rw(), potrf_buffer.write(), devInfo.write());
|
||||
// t.set_symbol("DPOTRF");
|
||||
t->*[&](cudaStream_t s, auto sAkk, auto buffer, auto info) {
|
||||
auto& h = get_cusolver_handle();
|
||||
cuda_safe_call(cusolverDnSetStream(h, s));
|
||||
|
||||
cuda_safe_call(cusolverDnDpotrf(
|
||||
h,
|
||||
uplo,
|
||||
sAkk.extent(0),
|
||||
sAkk.data_handle(),
|
||||
sAkk.stride(1),
|
||||
buffer.data_handle(),
|
||||
buffer.extent(0),
|
||||
info.data_handle()));
|
||||
};
|
||||
}
|
||||
|
||||
void DGEMM(
|
||||
cublasOperation_t transa,
|
||||
cublasOperation_t transb,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
int A_row,
|
||||
int A_col,
|
||||
class matrix<double>& B,
|
||||
int B_row,
|
||||
int B_col,
|
||||
double beta,
|
||||
class matrix<double>& C,
|
||||
int C_row,
|
||||
int C_col)
|
||||
{
|
||||
auto ignored = get_cublas_handle();
|
||||
auto t = ctx.task(A.handle(A_row, A_col).read(), B.handle(B_row, B_col).read(), C.handle(C_row, C_col).rw());
|
||||
// t.set_symbol("DGEMM");
|
||||
t->*[&](cudaStream_t s, auto sA, auto sB, auto sC) {
|
||||
auto& h = get_cublas_handle();
|
||||
cuda_safe_call(cublasSetStream(h, s));
|
||||
|
||||
auto k = (transa == CUBLAS_OP_N) ? sA.extent(1) : sA.extent(0);
|
||||
cuda_safe_call(cublasDgemm(
|
||||
h,
|
||||
transa,
|
||||
transb,
|
||||
sC.extent(0),
|
||||
sC.extent(1),
|
||||
k,
|
||||
&alpha,
|
||||
sA.data_handle(),
|
||||
sA.stride(1),
|
||||
sB.data_handle(),
|
||||
sB.stride(1),
|
||||
&beta,
|
||||
sC.data_handle(),
|
||||
sC.stride(1)));
|
||||
};
|
||||
}
|
||||
|
||||
void DSYRK(
|
||||
cublasFillMode_t uplo,
|
||||
cublasOperation_t trans,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
int A_row,
|
||||
int A_col,
|
||||
double beta,
|
||||
class matrix<double>& C,
|
||||
int C_row,
|
||||
int C_col)
|
||||
{
|
||||
auto ignored = get_cublas_handle();
|
||||
auto t = ctx.task(A.handle(A_row, A_col).read(), C.handle(C_row, C_col).rw());
|
||||
// t.set_symbol("DSYRK");
|
||||
t->*[&](cudaStream_t s, auto sA, auto sC) {
|
||||
auto& h = get_cublas_handle();
|
||||
cuda_safe_call(cublasSetStream(h, s));
|
||||
|
||||
// number of rows of matrix op(A) and C
|
||||
auto n = sC.extent(0);
|
||||
|
||||
// number of columns of matrix op(A)
|
||||
auto k = (trans == CUBLAS_OP_N) ? sA.extent(1) : sA.extent(0);
|
||||
|
||||
cuda_safe_call(
|
||||
cublasDsyrk(h, uplo, trans, n, k, &alpha, sA.data_handle(), sA.stride(1), &beta, sC.data_handle(), sC.stride(1)));
|
||||
};
|
||||
}
|
||||
|
||||
void DTRSM(
|
||||
cublasSideMode_t side,
|
||||
cublasFillMode_t uplo,
|
||||
cublasOperation_t transa,
|
||||
cublasDiagType_t diag,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
int A_row,
|
||||
int A_col,
|
||||
class matrix<double>& B,
|
||||
int B_row,
|
||||
int B_col)
|
||||
{
|
||||
auto ignored = get_cublas_handle();
|
||||
auto t = ctx.task(A.handle(A_row, A_col).read(), B.handle(B_row, B_col).rw());
|
||||
// t.set_symbol("DTRSM");
|
||||
t->*[&](cudaStream_t s, auto sA, auto sB) {
|
||||
auto& h = get_cublas_handle();
|
||||
cuda_safe_call(cublasSetStream(h, s));
|
||||
|
||||
cuda_safe_call(cublasDtrsm(
|
||||
h,
|
||||
side,
|
||||
uplo,
|
||||
transa,
|
||||
diag,
|
||||
sB.extent(0),
|
||||
sB.extent(1),
|
||||
&alpha,
|
||||
sA.data_handle(),
|
||||
sA.stride(1),
|
||||
sB.data_handle(),
|
||||
sB.stride(1)));
|
||||
};
|
||||
}
|
||||
|
||||
void PDNRM2_HOST(matrix<double>* A, double* result)
|
||||
{
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("red");
|
||||
#endif
|
||||
|
||||
for (size_t rowb = 0; rowb < A->mt; rowb++)
|
||||
{
|
||||
for (size_t colb = 0; colb < A->nt; colb++)
|
||||
{
|
||||
ctx.host_launch(A->handle(rowb, colb).read())->*[=](auto sA) {
|
||||
double res2 = 0.0;
|
||||
for (size_t col = 0; col < sA.extent(1); col++)
|
||||
{
|
||||
for (size_t row = 0; row < sA.extent(0); row++)
|
||||
{
|
||||
double v = sA(row, col);
|
||||
res2 += v * v;
|
||||
}
|
||||
}
|
||||
*result += res2;
|
||||
};
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
void PDPOTRF(matrix<double>& A)
|
||||
{
|
||||
nvtx_range r("PDPOTRF");
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("yellow");
|
||||
#endif
|
||||
|
||||
assert(A.m == A.n);
|
||||
assert(A.mt == A.nt);
|
||||
|
||||
int NBLOCKS = A.mt;
|
||||
assert(A.mb == A.nb);
|
||||
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
|
||||
for (int K = 0; K < NBLOCKS; K++)
|
||||
{
|
||||
int dev_akk = A.get_preferred_devid(K, K);
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(K, K)));
|
||||
DPOTRF(CUBLAS_FILL_MODE_LOWER, A, K, K);
|
||||
|
||||
for (int row = K + 1; row < NBLOCKS; row++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, K)));
|
||||
DTRSM(CUBLAS_SIDE_RIGHT, CUBLAS_FILL_MODE_LOWER, CUBLAS_OP_T, CUBLAS_DIAG_NON_UNIT, 1.0, A, K, K, A, row, K);
|
||||
|
||||
for (int col = K + 1; col < row; col++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, col)));
|
||||
DGEMM(CUBLAS_OP_N, CUBLAS_OP_T, -1.0, A, row, K, A, col, K, 1.0, A, row, col);
|
||||
}
|
||||
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, row)));
|
||||
DSYRK(CUBLAS_FILL_MODE_LOWER, CUBLAS_OP_N, -1.0, A, row, K, 1.0, A, row, row);
|
||||
}
|
||||
}
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
}
|
||||
|
||||
// Algorithm from PLASMA
|
||||
void PDTRSM(cublasSideMode_t side,
|
||||
cublasFillMode_t uplo,
|
||||
cublasOperation_t trans,
|
||||
cublasDiagType_t diag,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
class matrix<double>& B)
|
||||
{
|
||||
nvtx_range r("PDTRSM");
|
||||
|
||||
// std::cout << "[PDTRSM] START B MT " << B.mt << " NT " << B.nt << '\n';
|
||||
|
||||
if (side == CUBLAS_SIDE_LEFT)
|
||||
{
|
||||
if (uplo == CUBLAS_FILL_MODE_UPPER)
|
||||
{
|
||||
// TODO
|
||||
assert(0);
|
||||
abort();
|
||||
}
|
||||
else
|
||||
{
|
||||
//===========================================
|
||||
// CUBLAS_SIDE_LEFT / CUBLAS_FILL_MODE_LOWER / CUBLAS_OP_N
|
||||
//===========================================
|
||||
if (trans == CUBLAS_OP_N)
|
||||
{
|
||||
for (size_t k = 0; k < B.mt; k++)
|
||||
{
|
||||
double lalpha = k == 0 ? alpha : 1.0;
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(k, k)));
|
||||
DTRSM(side, uplo, trans, diag, lalpha, A, k, k, B, k, n);
|
||||
}
|
||||
for (size_t m = k + 1; m < B.mt; m++)
|
||||
{
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(m, k)));
|
||||
DGEMM(CUBLAS_OP_N, CUBLAS_OP_N, -1.0, A, m, k, B, k, n, lalpha, B, m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
//================================================
|
||||
// CUBLAS_SIDE_LEFT / CUBLAS_FILL_MODE_LOWER / CUBLAS_OP_[C|T]
|
||||
//================================================
|
||||
else
|
||||
{
|
||||
for (size_t k = 0; k < B.mt; k++)
|
||||
{
|
||||
double lalpha = k == 0 ? alpha : 1.0;
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(B.mt - k - 1, B.mt - k - 1)));
|
||||
DTRSM(side, uplo, trans, diag, lalpha, A, B.mt - k - 1, B.mt - k - 1, B, B.mt - k - 1, n);
|
||||
}
|
||||
for (size_t m = k + 1; m < B.mt; m++)
|
||||
{
|
||||
for (size_t n = 0; n < B.nt; n++)
|
||||
{
|
||||
cuda_safe_call(cudaSetDevice(A.get_preferred_devid(B.mt - k - 1, B.mt - 1 - m)));
|
||||
DGEMM(
|
||||
trans, CUBLAS_OP_N, -1.0, A, B.mt - k - 1, B.mt - 1 - m, B, B.mt - k - 1, n, lalpha, B, B.mt - 1 - m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
else
|
||||
{
|
||||
// TODO
|
||||
abort();
|
||||
}
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
// std::cout << "[PDTRSM] END" << '\n';
|
||||
}
|
||||
|
||||
void PDPOTRS(matrix<double>& A, class matrix<double>& B, cublasFillMode_t uplo)
|
||||
{
|
||||
nvtx_range r("PDPOTRS");
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("green");
|
||||
#endif
|
||||
|
||||
// std::cout << "[PDPOTRS] START" << '\n';
|
||||
// Call the parallel functions.
|
||||
PDTRSM(
|
||||
CUBLAS_SIDE_LEFT, uplo, uplo == CUBLAS_FILL_MODE_UPPER ? CUBLAS_OP_T : CUBLAS_OP_N, CUBLAS_DIAG_NON_UNIT, 1.0, A, B);
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("darkgreen");
|
||||
#endif
|
||||
|
||||
PDTRSM(
|
||||
CUBLAS_SIDE_LEFT, uplo, uplo == CUBLAS_FILL_MODE_UPPER ? CUBLAS_OP_N : CUBLAS_OP_T, CUBLAS_DIAG_NON_UNIT, 1.0, A, B);
|
||||
// std::cout << "[PDPOTRS] END" << '\n';
|
||||
}
|
||||
|
||||
/*****************************************************************************
|
||||
* Parallel tile matrix-matrix
|
||||
*multiplication.
|
||||
* @see plasma_omp_dgemm
|
||||
******************************************************************************/
|
||||
void PDGEMM(cublasOperation_t transa,
|
||||
cublasOperation_t transb,
|
||||
double alpha,
|
||||
class matrix<double>& A,
|
||||
class matrix<double>& B,
|
||||
double beta,
|
||||
class matrix<double>& C)
|
||||
{
|
||||
nvtx_range r("PDGEMM");
|
||||
|
||||
#ifdef HAVE_DOT
|
||||
reserved::dot::set_current_color("blue");
|
||||
#endif
|
||||
|
||||
for (size_t m = 0; m < C.mt; m++)
|
||||
{
|
||||
for (size_t n = 0; n < C.nt; n++)
|
||||
{
|
||||
//=========================================
|
||||
// alpha*A*B does not contribute; scale C
|
||||
//=========================================
|
||||
int inner_k = transa == CUBLAS_OP_N ? A.n : A.m;
|
||||
if (alpha == 0.0 || inner_k == 0)
|
||||
{
|
||||
DGEMM(transa, transb, alpha, A, 0, 0, B, 0, 0, beta, C, m, n);
|
||||
}
|
||||
else if (transa == CUBLAS_OP_N)
|
||||
{
|
||||
//================================
|
||||
// CUBLAS_OP_N / CUBLAS_OP_N
|
||||
//================================
|
||||
if (transb == CUBLAS_OP_N)
|
||||
{
|
||||
for (size_t k = 0; k < A.nt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, m, k, B, k, n, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
//=====================================
|
||||
// CUBLAS_OP_N / CUBLAS_OP_T
|
||||
//=====================================
|
||||
else
|
||||
{
|
||||
for (size_t k = 0; k < A.nt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, m, k, B, n, k, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
else
|
||||
{
|
||||
//=====================================
|
||||
// CUBLAS_OP_T / CUBLAS_OP_N
|
||||
//=====================================
|
||||
if (transb == CUBLAS_OP_N)
|
||||
{
|
||||
for (size_t k = 0; k < A.mt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, k, m, B, k, n, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
//==========================================
|
||||
// CUBLAS_OP_T / CUBLAS_OP_T
|
||||
//==========================================
|
||||
else
|
||||
{
|
||||
for (size_t k = 0; k < A.mt; k++)
|
||||
{
|
||||
double zbeta = k == 0 ? beta : 1.0;
|
||||
DGEMM(transa, transb, alpha, A, k, m, B, n, k, zbeta, C, m, n);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
int N = 1024;
|
||||
int NB = 128;
|
||||
|
||||
if (argc > 1)
|
||||
{
|
||||
N = atoi(argv[1]);
|
||||
}
|
||||
|
||||
if (argc > 2)
|
||||
{
|
||||
NB = atoi(argv[2]);
|
||||
}
|
||||
|
||||
assert(N % NB == 0);
|
||||
|
||||
// Set up CUBLAS and CUSOLVER
|
||||
int ndevs;
|
||||
cuda_safe_call(cudaGetDeviceCount(&ndevs));
|
||||
|
||||
cuda_safe_call(cudaSetDevice(0));
|
||||
|
||||
matrix<double> A(N, N, NB, NB, true, "A");
|
||||
matrix<double> Aref(N, N, NB, NB, false, "Aref");
|
||||
|
||||
// (Hilbert matrix + 2*N*Id) to have a diagonal dominant matrix
|
||||
auto hilbert = [](matrix<double>& mat, int row, int col) {
|
||||
return 1.0 / (col + row + 1.0) + 2.0 * mat.n * (col == row);
|
||||
};
|
||||
|
||||
Aref.fill(hilbert);
|
||||
A.fill(hilbert);
|
||||
|
||||
/* Right-hand side */
|
||||
matrix<double> B_potrs(N, 1, NB, 1, false, "B");
|
||||
matrix<double> Bref_potrs(N, 1, NB, 1, false, "Bref");
|
||||
|
||||
auto rhs_vals = [](matrix<double>& /*unused*/, int row, int /*unused*/) {
|
||||
return 1.0 * (row + 1);
|
||||
};
|
||||
B_potrs.fill(rhs_vals);
|
||||
Bref_potrs.fill(rhs_vals);
|
||||
|
||||
int check_result = 1;
|
||||
if (getenv("CHECK_RESULT"))
|
||||
{
|
||||
check_result = atoi(getenv("CHECK_RESULT"));
|
||||
}
|
||||
|
||||
// // Compute ||Bref||
|
||||
double Bref_nrm2 = 0.0;
|
||||
double res_nrm2 = 0.0;
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
PDNRM2_HOST(&Bref_potrs, &Bref_nrm2);
|
||||
}
|
||||
|
||||
// for (size_t row = 0; row < A.mt; row++)
|
||||
// {
|
||||
// for (size_t col = 0; col <= row; col++)
|
||||
// {
|
||||
// cuda_safe_call(cudaSetDevice(A.get_preferred_devid(row, col)));
|
||||
// NOOP(A, row, col);
|
||||
// }
|
||||
// }
|
||||
|
||||
PDPOTRF(A);
|
||||
|
||||
/*
|
||||
* POTRS
|
||||
*/
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
// Solve AX = B and put the result in B
|
||||
PDPOTRS(A, B_potrs, CUBLAS_FILL_MODE_LOWER);
|
||||
|
||||
// Compute (AX - B)
|
||||
// Bref = (Aref*B - Bref)
|
||||
PDGEMM(CUBLAS_OP_N, CUBLAS_OP_N, 1.0, Aref, B_potrs, -1.0, Bref_potrs);
|
||||
|
||||
// Compute ||AX - B|| = ||Bref||
|
||||
PDNRM2_HOST(&Bref_potrs, &res_nrm2);
|
||||
}
|
||||
|
||||
ctx.finalize();
|
||||
|
||||
if (check_result)
|
||||
{
|
||||
double residual = sqrt(res_nrm2) / sqrt(Bref_nrm2);
|
||||
// std::cout << "[POTRS] ||AX - B|| : " << sqrt(res_nrm2) << '\n';
|
||||
// std::cout << "[POTRS] ||B|| : " << sqrt(Bref_nrm2) << '\n';
|
||||
// std::cout << "[POTRS] RESIDUAL (||AX - B||/||B||) : " << residual << '\n';
|
||||
assert(residual < 0.01);
|
||||
}
|
||||
|
||||
return 0;
|
||||
}
|
||||
@@ -1,129 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/stf.cuh>
|
||||
|
||||
#include <random>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
struct body
|
||||
{
|
||||
// mass
|
||||
double mass;
|
||||
// position
|
||||
double pos[3];
|
||||
// speed
|
||||
double vel[3];
|
||||
// acceleration
|
||||
double acc[3];
|
||||
};
|
||||
|
||||
int main()
|
||||
{
|
||||
constexpr double kSofteningSquared = 1e-3;
|
||||
constexpr double kG = 6.67259e-11;
|
||||
|
||||
size_t BODY_CNT = 4096;
|
||||
|
||||
double dt = 0.1;
|
||||
size_t NITER = 25;
|
||||
|
||||
context ctx;
|
||||
|
||||
std::vector<body> particles;
|
||||
particles.resize(BODY_CNT);
|
||||
|
||||
// Initialize particles
|
||||
std::random_device rd;
|
||||
std::mt19937 gen(rd());
|
||||
std::uniform_real_distribution<> dis(-1.0, 1.0);
|
||||
for (auto& p : particles)
|
||||
{
|
||||
p.mass = 1.0;
|
||||
|
||||
p.pos[0] = dis(gen);
|
||||
p.pos[1] = dis(gen);
|
||||
p.pos[2] = dis(gen);
|
||||
|
||||
p.vel[0] = dis(gen);
|
||||
p.vel[1] = dis(gen);
|
||||
p.vel[2] = dis(gen);
|
||||
|
||||
p.acc[0] = 0.0;
|
||||
p.acc[1] = 0.0;
|
||||
p.acc[2] = 0.0;
|
||||
}
|
||||
|
||||
auto h_particles = ctx.logical_data(make_slice(&particles[0], BODY_CNT));
|
||||
|
||||
auto fn = [dt](context ctx, logical_data<slice<body>> h_particles) {
|
||||
// Compute accelerations
|
||||
ctx.parallel_for(h_particles.shape(), h_particles.rw())->*[=] _CCCL_DEVICE __host__(size_t i, slice<body> p) {
|
||||
double acc[3];
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
acc[k] = p(i).acc[k];
|
||||
}
|
||||
|
||||
for (size_t j = 0; j < p.extent(0); j++)
|
||||
{
|
||||
if (i != j)
|
||||
{
|
||||
double d[3];
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
d[k] = p(j).pos[k] - p(i).pos[k];
|
||||
}
|
||||
|
||||
double dist = d[0] * d[0] + d[1] * d[1] + d[2] * d[2] + kSofteningSquared;
|
||||
double dist_inv = 1.0 / sqrt(dist);
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
acc[k] += d[k] * kG * p(j).mass * dist_inv * dist_inv * dist_inv;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).acc[k] = acc[k];
|
||||
}
|
||||
};
|
||||
|
||||
// Update velocity and positions
|
||||
ctx.parallel_for(h_particles.shape(), h_particles.rw())->*[=] __host__ __device__(size_t i, slice<body> p) {
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).vel[k] += p(i).acc[k] * dt;
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).pos[k] += p(i).vel[k] * dt;
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).acc[k] = 0.0;
|
||||
}
|
||||
};
|
||||
};
|
||||
|
||||
algorithm one_iter;
|
||||
for (size_t iter = 0; iter < NITER; iter++)
|
||||
{
|
||||
// fprintf(stderr, "ITER %ld\n", iter);
|
||||
one_iter.run_as_task(fn, ctx, h_particles.rw());
|
||||
}
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,283 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/stf.cuh>
|
||||
|
||||
#include <random>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
struct body
|
||||
{
|
||||
// mass
|
||||
double mass;
|
||||
// position
|
||||
double pos[3];
|
||||
// speed
|
||||
double vel[3];
|
||||
};
|
||||
|
||||
// Function to write VTK file for a single time step
|
||||
void writeVTKFile(context& ctx,
|
||||
const std::string& filename,
|
||||
size_t BLOCK_SIZE,
|
||||
size_t BODY_CNT,
|
||||
std::vector<logical_data<slice<body>>> parts)
|
||||
{
|
||||
std::ofstream outfile(filename);
|
||||
|
||||
if (!outfile)
|
||||
{
|
||||
std::cerr << "Error opening file: " << filename << '\n';
|
||||
return;
|
||||
}
|
||||
|
||||
outfile << "# vtk DataFile Version 4.2\n";
|
||||
outfile << "Position Data\n";
|
||||
outfile << "ASCII\n";
|
||||
outfile << "DATASET UNSTRUCTURED_GRID\n";
|
||||
outfile << "POINTS " << BODY_CNT << " float\n";
|
||||
|
||||
std::vector<double> dump(3 * BODY_CNT);
|
||||
|
||||
for (size_t b = 0; b < parts.size(); b++)
|
||||
{
|
||||
ctx.task(exec_place::host(), parts[b].read())->*[&](cudaStream_t s, slice<const body> p) {
|
||||
cuda_safe_call(cudaStreamSynchronize(s));
|
||||
for (size_t i = 0; i < p.size(); i++)
|
||||
{
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
dump[3 * (i + b * BLOCK_SIZE) + k] = p(i).pos[k];
|
||||
}
|
||||
}
|
||||
};
|
||||
}
|
||||
|
||||
for (size_t p = 0; p < BODY_CNT; p++)
|
||||
{
|
||||
outfile << dump[3 * p] << " " << dump[3 * p + 1] << " " << dump[3 * p + 2] << "\n";
|
||||
}
|
||||
|
||||
outfile.close();
|
||||
}
|
||||
|
||||
void load_input_file(std::string filename, std::vector<body>& particles)
|
||||
{
|
||||
std::ifstream infile(filename);
|
||||
if (!infile)
|
||||
{
|
||||
std::cerr << "Error opening file." << '\n';
|
||||
abort();
|
||||
return;
|
||||
}
|
||||
|
||||
double mass, posX, posY, posZ, velX, velY, velZ;
|
||||
size_t ind = 0;
|
||||
|
||||
// Loop until we reach the end of the file
|
||||
while (infile >> mass >> posX >> posY >> posZ >> velX >> velY >> velZ)
|
||||
{
|
||||
body p;
|
||||
p.mass = mass;
|
||||
|
||||
p.pos[0] = posX;
|
||||
p.pos[1] = posY;
|
||||
p.pos[2] = posZ;
|
||||
|
||||
p.vel[0] = velX;
|
||||
p.vel[1] = velY;
|
||||
p.vel[2] = velZ;
|
||||
|
||||
// // Display first bodies
|
||||
// if (ind < 10) {
|
||||
// fprintf(stderr, "body xyz %e %e %e dxyz %e %e %e m %e\n", p.pos[0], p.pos[1], p.pos[2], p.vel[0],
|
||||
// p.vel[1], p.vel[2], p.mass);
|
||||
//}
|
||||
|
||||
ind++;
|
||||
particles.push_back(p);
|
||||
}
|
||||
|
||||
fprintf(stderr, "Loaded %zu bodies from %s...\n", ind, filename.c_str());
|
||||
}
|
||||
|
||||
int main(int argc, char** argv)
|
||||
{
|
||||
constexpr double kSofteningSquared = 1e-9;
|
||||
// constexpr double kG = 6.67259e-11;
|
||||
constexpr double kG = 1.0;
|
||||
|
||||
size_t BODY_CNT = 128ULL * 1024ULL;
|
||||
size_t BLOCK_SIZE = 16 * 1024ULL;
|
||||
|
||||
std::vector<body> particles;
|
||||
|
||||
// Initialize particles
|
||||
if (argc > 1)
|
||||
{
|
||||
// Get dataset from file
|
||||
std::string filename = argv[1];
|
||||
|
||||
load_input_file(filename, particles);
|
||||
|
||||
BODY_CNT = particles.size();
|
||||
BLOCK_SIZE = (BODY_CNT + 7) / 8;
|
||||
}
|
||||
else
|
||||
{
|
||||
// Random distribution
|
||||
BODY_CNT = 32ULL * 1024ULL;
|
||||
particles.resize(BODY_CNT);
|
||||
|
||||
std::random_device rd;
|
||||
std::mt19937 gen(rd());
|
||||
std::uniform_real_distribution<> dis(-1.0, 1.0);
|
||||
for (auto& p : particles)
|
||||
{
|
||||
p.mass = 1.0;
|
||||
|
||||
p.pos[0] = dis(gen);
|
||||
p.pos[1] = dis(gen);
|
||||
p.pos[2] = dis(gen);
|
||||
|
||||
p.vel[0] = dis(gen);
|
||||
p.vel[1] = dis(gen);
|
||||
p.vel[2] = dis(gen);
|
||||
}
|
||||
}
|
||||
|
||||
cuda_safe_call(cudaHostRegister(&particles[0], BODY_CNT * sizeof(body), cudaHostRegisterPortable));
|
||||
|
||||
double dt = 0.005;
|
||||
size_t NITER = 7; // 7000;
|
||||
|
||||
context ctx;
|
||||
|
||||
std::vector<logical_data<slice<body>>> parts;
|
||||
|
||||
// Accelerations
|
||||
std::vector<logical_data<slice<double, 2>>> acc_parts;
|
||||
|
||||
size_t block_cnt = (BODY_CNT + BLOCK_SIZE - 1) / BLOCK_SIZE;
|
||||
for (size_t i = 0; i < block_cnt; i++)
|
||||
{
|
||||
size_t first = i * BLOCK_SIZE;
|
||||
size_t last = std::min((i + 1) * BLOCK_SIZE, BODY_CNT);
|
||||
auto p_i = ctx.logical_data(make_slice(&particles[first], last - first));
|
||||
parts.push_back(p_i);
|
||||
|
||||
auto acc_p_i = ctx.logical_data(shape_of<slice<double, 2>>(last - first, 3));
|
||||
acc_parts.push_back(acc_p_i);
|
||||
}
|
||||
|
||||
int ngpus;
|
||||
cuda_safe_call(cudaGetDeviceCount(&ngpus));
|
||||
|
||||
cudaEvent_t start;
|
||||
cuda_safe_call(cudaEventCreate(&start));
|
||||
cuda_safe_call(cudaEventRecord(start, ctx.fence()));
|
||||
|
||||
for (size_t iter = 0; iter < NITER; iter++)
|
||||
{
|
||||
// Initialize acceleration to 0
|
||||
for (size_t b = 0; b < block_cnt; b++)
|
||||
{
|
||||
ctx.launch(exec_place::device(b % ngpus), acc_parts[b].write())
|
||||
//.set_symbol("init_acc")
|
||||
->*[=] _CCCL_DEVICE(auto t, slice<double, 2> acc) {
|
||||
for (size_t i = t.rank(); i < acc.extent(0); i += t.size())
|
||||
{
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
acc(i, k) = 0.0;
|
||||
}
|
||||
}
|
||||
};
|
||||
}
|
||||
|
||||
// Compute accelerations
|
||||
for (size_t b = 0; b < block_cnt; b++)
|
||||
{
|
||||
for (size_t b_other = 0; b_other < block_cnt; b_other++)
|
||||
{
|
||||
ctx.launch(exec_place::device(b % ngpus), parts[b].read(), parts[b_other].read(), acc_parts[b].rw())
|
||||
//.set_symbol("compute_acc")
|
||||
->*[=] _CCCL_DEVICE(auto t, slice<const body> p, slice<const body> p_other, slice<double, 2> acc) {
|
||||
for (size_t i = t.rank(); i < p.extent(0); i += t.size())
|
||||
{
|
||||
for (size_t j = 0; j < p_other.extent(0); j++)
|
||||
{
|
||||
if ((b * BLOCK_SIZE + i) != (b_other * BLOCK_SIZE + j))
|
||||
{
|
||||
double d[3];
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
d[k] = p_other(j).pos[k] - p(i).pos[k];
|
||||
}
|
||||
|
||||
double dist = d[0] * d[0] + d[1] * d[1] + d[2] * d[2] + kSofteningSquared;
|
||||
double dist_inv = 1.0 / sqrt(dist);
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
acc(i, k) += d[k] * kG * p_other(j).mass * dist_inv * dist_inv * dist_inv;
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
};
|
||||
}
|
||||
}
|
||||
|
||||
for (size_t b = 0; b < block_cnt; b++)
|
||||
{
|
||||
// Update velocity and positions
|
||||
ctx.launch(exec_place::device(b % ngpus), parts[b].rw(), acc_parts[b].read())
|
||||
//.set_symbol("update")
|
||||
->*[=] _CCCL_DEVICE(auto t, slice<body> p, slice<const double, 2> acc) {
|
||||
for (size_t i = t.rank(); i < p.extent(0); i += t.size())
|
||||
{
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).vel[k] += acc(i, k) * dt;
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).pos[k] += p(i).vel[k] * dt;
|
||||
}
|
||||
}
|
||||
};
|
||||
}
|
||||
|
||||
// Write the VTK file for this time step
|
||||
const char* dump_freq_str = getenv("DUMP_FREQ");
|
||||
if (dump_freq_str && iter % atoi(dump_freq_str) == 0)
|
||||
{
|
||||
std::string filename = "time_step_" + std::to_string(iter) + ".vtk";
|
||||
writeVTKFile(ctx, filename, BLOCK_SIZE, BODY_CNT, parts);
|
||||
}
|
||||
}
|
||||
|
||||
cudaEvent_t stop;
|
||||
cuda_safe_call(cudaEventCreate(&stop));
|
||||
cuda_safe_call(cudaEventRecord(stop, ctx.fence()));
|
||||
|
||||
ctx.finalize();
|
||||
|
||||
float elapsed;
|
||||
cuda_safe_call(cudaEventElapsedTime(&elapsed, start, stop));
|
||||
|
||||
// rough approximation !
|
||||
double FLOP_COUNT = 21.0 * (1.0 * BODY_CNT) * (1.0 * BODY_CNT) * NITER;
|
||||
|
||||
printf("NBODY: elapsed %f ms, %f GFLOPS\n", elapsed, FLOP_COUNT / elapsed / 1000000.0);
|
||||
}
|
||||
@@ -1,122 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
#include <cuda/experimental/stf.cuh>
|
||||
|
||||
#include <random>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
struct body
|
||||
{
|
||||
// mass
|
||||
double mass;
|
||||
// position
|
||||
double pos[3];
|
||||
// speed
|
||||
double vel[3];
|
||||
// acceleration
|
||||
double acc[3];
|
||||
};
|
||||
|
||||
int main()
|
||||
{
|
||||
constexpr double kSofteningSquared = 1e-3;
|
||||
constexpr double kG = 6.67259e-11;
|
||||
|
||||
size_t BODY_CNT = 4096;
|
||||
|
||||
double dt = 0.1;
|
||||
size_t NITER = 25;
|
||||
|
||||
context ctx = graph_ctx();
|
||||
|
||||
std::vector<body> particles;
|
||||
particles.resize(BODY_CNT);
|
||||
|
||||
// Initialize particles
|
||||
std::random_device rd;
|
||||
std::mt19937 gen(rd());
|
||||
std::uniform_real_distribution<> dis(-1.0, 1.0);
|
||||
for (auto& p : particles)
|
||||
{
|
||||
p.mass = 1.0;
|
||||
|
||||
p.pos[0] = dis(gen);
|
||||
p.pos[1] = dis(gen);
|
||||
p.pos[2] = dis(gen);
|
||||
|
||||
p.vel[0] = dis(gen);
|
||||
p.vel[1] = dis(gen);
|
||||
p.vel[2] = dis(gen);
|
||||
|
||||
p.acc[0] = 0.0;
|
||||
p.acc[1] = 0.0;
|
||||
p.acc[2] = 0.0;
|
||||
}
|
||||
|
||||
auto h_particles = ctx.logical_data(make_slice(&particles[0], BODY_CNT));
|
||||
|
||||
ctx.repeat(NITER)->*[&](context ctx, size_t) {
|
||||
// Compute accelerations
|
||||
ctx.parallel_for(h_particles.shape(), h_particles.rw())->*[=] _CCCL_DEVICE __host__(size_t i, slice<body> p) {
|
||||
double acc[3];
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
acc[k] = p(i).acc[k];
|
||||
}
|
||||
|
||||
for (size_t j = 0; j < p.extent(0); j++)
|
||||
{
|
||||
if (i != j)
|
||||
{
|
||||
double d[3];
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
d[k] = p(j).pos[k] - p(i).pos[k];
|
||||
}
|
||||
|
||||
double dist = d[0] * d[0] + d[1] * d[1] + d[2] * d[2] + kSofteningSquared;
|
||||
double dist_inv = 1.0 / sqrt(dist);
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
acc[k] += d[k] * kG * p(j).mass * dist_inv * dist_inv * dist_inv;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).acc[k] = acc[k];
|
||||
}
|
||||
};
|
||||
|
||||
// Update velocity and positions
|
||||
ctx.parallel_for(h_particles.shape(), h_particles.rw())->*[=] __host__ __device__(size_t i, slice<body> p) {
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).vel[k] += p(i).acc[k] * dt;
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).pos[k] += p(i).vel[k] * dt;
|
||||
}
|
||||
|
||||
for (size_t k = 0; k < 3; k++)
|
||||
{
|
||||
p(i).acc[k] = 0.0;
|
||||
}
|
||||
};
|
||||
};
|
||||
|
||||
ctx.finalize();
|
||||
}
|
||||
@@ -1,123 +0,0 @@
|
||||
//===----------------------------------------------------------------------===//
|
||||
//
|
||||
// Part of CUDASTF in CUDA C++ Core Libraries,
|
||||
// under the Apache License v2.0 with LLVM Exceptions.
|
||||
// See https://llvm.org/LICENSE.txt for license information.
|
||||
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
|
||||
// SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES.
|
||||
//
|
||||
//===----------------------------------------------------------------------===//
|
||||
|
||||
/**
|
||||
* @file
|
||||
*
|
||||
* @brief Test that the cuda_kernel construct works with global kernels, CUfunction and CUkernel entries.
|
||||
*
|
||||
*/
|
||||
|
||||
#include <cuda/experimental/stf.cuh>
|
||||
|
||||
using namespace cuda::experimental::stf;
|
||||
|
||||
__global__ void axpy(double a, slice<const double> x, slice<double> y)
|
||||
{
|
||||
int tid = blockIdx.x * blockDim.x + threadIdx.x;
|
||||
int nthreads = gridDim.x * blockDim.x;
|
||||
|
||||
for (int i = tid; i < x.size(); i += nthreads)
|
||||
{
|
||||
y(i) += a * x(i);
|
||||
}
|
||||
}
|
||||
|
||||
double X0(int i)
|
||||
{
|
||||
return sin((double) i);
|
||||
}
|
||||
|
||||
double Y0(int i)
|
||||
{
|
||||
return cos((double) i);
|
||||
}
|
||||
|
||||
void test(bool is_graph)
|
||||
{
|
||||
context ctx;
|
||||
if (is_graph)
|
||||
{
|
||||
ctx = graph_ctx();
|
||||
}
|
||||
|
||||
const size_t N = 16;
|
||||
double X[N], Y[N];
|
||||
|
||||
for (size_t i = 0; i < N; i++)
|
||||
{
|
||||
X[i] = X0(i);
|
||||
Y[i] = Y0(i);
|
||||
}
|
||||
|
||||
// Number of times we have applied the axpy kernel
|
||||
int num_axpy = 0;
|
||||
|
||||
double alpha = 3.14;
|
||||
|
||||
auto lX = ctx.logical_data(X);
|
||||
auto lY = ctx.logical_data(Y);
|
||||
|
||||
// runtime global kernel
|
||||
ctx.cuda_kernel(lX.read(), lY.rw())->*[&](auto dX, auto dY) {
|
||||
// axpy<<<16, 128, 0, ...>>>(alpha, dX, dY)
|
||||
return cuda_kernel_desc{axpy, 16, 128, 0, alpha, dX, dY};
|
||||
};
|
||||
num_axpy++;
|
||||
|
||||
// CUfunction driver API
|
||||
CUfunction axpy_fun;
|
||||
cuda_safe_call(cudaGetFuncBySymbol(&axpy_fun, (void*) axpy));
|
||||
|
||||
ctx.cuda_kernel(lX.read(), lY.rw())->*[&](auto dX, auto dY) {
|
||||
return cuda_kernel_desc{axpy_fun, 16, 128, 0, alpha, dX, dY};
|
||||
};
|
||||
num_axpy++;
|
||||
|
||||
#if _CCCL_CTK_AT_LEAST(12, 1)
|
||||
// CUkernel driver API
|
||||
CUkernel axpy_kernel;
|
||||
cuda_safe_call(cudaGetKernel(&axpy_kernel, (void*) axpy));
|
||||
|
||||
ctx.cuda_kernel(lX.read(), lY.rw())->*[&](auto dX, auto dY) {
|
||||
return cuda_kernel_desc{axpy_kernel, 16, 128, 0, alpha, dX, dY};
|
||||
};
|
||||
num_axpy++;
|
||||
#endif
|
||||
|
||||
/* Some extra sanity checks, we put this in a dummy task to get access to dX and dY values */
|
||||
ctx.task(lX.read(), lY.rw())->*[&](auto, auto dX, auto dY) {
|
||||
int nregs = cuda_kernel_desc{axpy, 16, 128, 0, alpha, dX, dY}.get_num_registers();
|
||||
|
||||
int nregs_fun = cuda_kernel_desc{axpy_fun, 16, 128, 0, alpha, dX, dY}.get_num_registers();
|
||||
_CCCL_ASSERT(nregs == nregs_fun, "invalid value");
|
||||
|
||||
#if _CCCL_CTK_AT_LEAST(12, 1)
|
||||
int nregs_kernel = cuda_kernel_desc{axpy_kernel, 16, 128, 0, alpha, dX, dY}.get_num_registers();
|
||||
_CCCL_ASSERT(nregs == nregs_kernel, "invalid value");
|
||||
#endif
|
||||
};
|
||||
|
||||
ctx.finalize();
|
||||
|
||||
for (size_t i = 0; i < N; i++)
|
||||
{
|
||||
_CCCL_ASSERT(fabs(Y[i] - (Y0(i) + num_axpy * alpha * X0(i))) < 0.0001, "Invalid result");
|
||||
_CCCL_ASSERT(fabs(X[i] - X0(i)) < 0.0001, "Invalid result");
|
||||
}
|
||||
}
|
||||
|
||||
int main()
|
||||
{
|
||||
// stream context
|
||||
test(false);
|
||||
// graph context
|
||||
test(true);
|
||||
}
|
||||
Reference in New Issue
Block a user